#include "Utils.h"
#include "LikelihoodDynamicLibrary.h"

LikelihoodDynamicLibrary::LikelihoodDynamicLibrary(size_t numthreads)
	: numthreads(numthreads)
	, allocated(false)
	, allocate_models(NULL)
	, deallocate_models(NULL)
	, evaluate_likelihood(NULL)
	, get_output(NULL)
#if PLATFORM_WINDOWS
	, module(NULL)
#else
	, library(NULL)
#endif
{
}

LikelihoodDynamicLibrary::~LikelihoodDynamicLibrary()
{
	if (allocated && deallocate_models) {
		deallocate_models();
	}

#if PLATFORM_WINDOWS
	if (module) {
		FreeLibrary(module);
	}
#else
	if (library) {
		dlclose(library);
	}
#endif
}

bool LikelihoodDynamicLibrary::Load(const std::string& filename)
{
#if PLATFORM_WINDOWS
	if (module != NULL) {
		LOGERROR("Library already loaded");
		return false;
	}

	module = LoadLibrary(filename.c_str());
	if (!module) {
		LOGERROR("LoadLibrary(\"%s\") failed: %u", filename.c_str(), GetLastError());
		return false;
	}
#else
	dlerror();
	library = dlopen(filename.c_str(), RTLD_NOW);
	if (!library) {
		LOGERROR("dlopen(\"%s\") failed: %s", filename.c_str(), dlerror());
		return false;
	}
#endif

	allocate_models = GetFunction<AllocateModelsFn>("AllocateModels");
	deallocate_models = GetFunction<DeallocateModelsFn>("DeallocateModels");
	evaluate_likelihood = GetFunction<EvaluateLikelihoodFn>("EvaluateLikelihood");
	get_output = GetFunction<GetOutputFn>("GetOutput");

	if (!allocate_models || !deallocate_models || !evaluate_likelihood || !get_output) {
		return false;
	}

	if (!allocate_models(numthreads)) {
		LOGERROR("AllocateModels failed");
		return false;
	}
	allocated = true;

	return true;
}

bool LikelihoodDynamicLibrary::EvaluateLogProbability(size_t threadix, const Real* values, Real& logp)
{
	evaluate_likelihood(threadix, values, logp);
	return true;
}

bool LikelihoodDynamicLibrary::GetOutput(size_t threadix, Real* output_values, size_t& numvalues)
{
	get_output(threadix, output_values, numvalues);
	return true;
}
