#include "Utils.h"
#include "CVODESolver.h"

#include <cvode/cvode.h>
#include <cvode/cvode_dense.h>
#include <cvode/cvode_impl.h>
#include <nvector/nvector_serial.h>

static const size_t MAX_CVODE_STEPS = 2000;

static void static_cvode_err_fn(int error_code, const char *module, const char *function, char *msg, void *user_data)
{
#if 0
	LOGERROR("CVODE error %d in module %s, function %s: %s", error_code, module, function, msg);
#endif
}

static int static_cvode_rhs_fn(OdeReal t, N_Vector y, N_Vector ydot, void* user_data)
{
	CVODESolver* solver = reinterpret_cast<CVODESolver*>(user_data);
	return solver->cvode_rhs_fn(t, NV_DATA_S(y), NV_DATA_S(ydot));
}

static int static_cvode_jac_fn(long int N, OdeReal t, N_Vector y, N_Vector fy, DlsMat Jac, void* user_data, N_Vector ytmp1, N_Vector ytmp2, N_Vector ytmp3)
{
	CVODESolver* solver = reinterpret_cast<CVODESolver*>(user_data);
	return solver->cvode_jac_fn(t, NV_DATA_S(y), NV_DATA_S(fy), Jac->cols);
}

static int solver_init(CVodeMem cv_mem)
{
	CVODESolver* solver = reinterpret_cast<CVODESolver*>(cv_mem->cv_lmem);
	return solver->cvode_solver_init();
}

static int solver_setup(CVodeMem cv_mem, int convfail, N_Vector ypred, N_Vector fpred, booleantype *jcurPtr, N_Vector vtemp1, N_Vector vtemp2, N_Vector vtemp3)
{
	CVODESolver* solver = reinterpret_cast<CVODESolver*>(cv_mem->cv_lmem);
	return solver->cvode_solver_setup(cv_mem, convfail, NV_DATA_S(ypred), jcurPtr);
}

static int solver_solve(CVodeMem cv_mem, N_Vector b, N_Vector weight, N_Vector ycur, N_Vector fcur)
{
	CVODESolver* solver = reinterpret_cast<CVODESolver*>(cv_mem->cv_lmem);
	return solver->cvode_solver_solve(cv_mem, NV_DATA_S(b));
}

static void solver_free(CVodeMem cv_mem)
{
	CVODESolver* solver = reinterpret_cast<CVODESolver*>(cv_mem->cv_lmem);
	solver->cvode_solver_free();
}

CVODESolver::CVODESolver()
	: cvode_mem(NULL)
	, y(NULL)
	, interpolate_y(NULL)
	, user_data(NULL)
	, linear_solver(NULL)
{
}

CVODESolver::~CVODESolver()
{
	if (y) {
		N_VDestroy_Serial((N_Vector)y);
	}
	if (interpolate_y) {
		N_VDestroy_Serial((N_Vector)interpolate_y);
	}
}

void CVODESolver::SetLinearSolver(LinearSolver* solver)
{
	linear_solver = solver;
}

bool CVODESolver::Initialize(size_t N, void* user)
{
	if (N > std::numeric_limits<long>::max()) {
		LOGERROR("System too large");
		return false;
	}
	this->N = N;

	y = N_VNew_Serial((long)N);
	interpolate_y = N_VNew_Serial((long)N);

	cvode_mem = CVodeCreate(CV_BDF, CV_NEWTON);
	//cvode_mem = CVodeCreate(CV_ADAMS, CV_FUNCTIONAL);
	if (!cvode_mem) {
		LOGERROR("CVodeCreate failed");
		return false;
	}

	user_data = user;

	CVodeInit(cvode_mem, &static_cvode_rhs_fn, 0.0, (N_Vector)y);
	CVodeSetUserData(cvode_mem, this);
	CVodeSStolerances(cvode_mem, (OdeReal)1e-6, (OdeReal)1e-6);
	CVodeSetMaxNumSteps(cvode_mem, 500); // Note that this is not actually used, since we take one step at a time.
	CVodeSetErrHandlerFn(cvode_mem, &static_cvode_err_fn, this);
	CVodeSetMinStep(cvode_mem, (OdeReal)1e-12);
	if (linear_solver) {
		CVodeMem cv_mem = (CVodeMem)cvode_mem;
		cv_mem->cv_linit = &solver_init;
		cv_mem->cv_lsetup = &solver_setup;
		cv_mem->cv_lsolve = &solver_solve;
		cv_mem->cv_lfree = &solver_free;
		cv_mem->cv_lmem = this;
		cv_mem->cv_setupNonNull = 1;
	} else {
		CVDense(cvode_mem, (long)N);
		if (jacobian) {
			CVDlsSetDenseJacFn(cvode_mem, &static_cvode_jac_fn);
		}
	}

	step_counts.resize(MAX_CVODE_STEPS + 1, 0);
	failed_step_counts.resize(MAX_CVODE_STEPS + 1, 0);

	return true;
}

bool CVODESolver::SetTolerance(Real relative, Real absolute)
{
	CVodeSStolerances(cvode_mem, (OdeReal)relative, (OdeReal)absolute);
	return true;
}

bool CVODESolver::SetTolerance(Real relative, VectorReal absolute)
{
	if (absolute.size() != N) {
		return false;
	}

	N_Vector abstol = N_VNew_Serial((long)N); 
	for (size_t i = 0; i < N; i++) {
		NV_Ith_S(abstol,i) = (OdeReal)absolute(i);
	}
	CVodeSVtolerances(cvode_mem, (OdeReal)relative, abstol);
	N_VDestroy_Serial(abstol);
	return true;
}

void CVODESolver::SetUserData(void* user)
{
	user_data = user;
}

void CVODESolver::SetDerivativeFunction(TDeriviativeFunction f)
{
	derivative = f;
}

void CVODESolver::SetJacobianFunction(TJacobianFunction f)
{
	jacobian = f;
}

bool CVODESolver::Simulate(const Real* initial_conditions, const std::vector<Real>& timepoints, MatrixReal& output)
{
	if (timepoints.empty()) {
		LOGERROR("No time points requested");
		return false;
	}

	for (size_t i = 0; i < N; i++) {
		NV_Ith_S(((N_Vector)y),i) = (OdeReal)initial_conditions[i];
	}

	CVodeReInit(cvode_mem, 0.0, (N_Vector)y);
	output.setZero(N, timepoints.size());

	// If the first time point is at t=0, copy the initial conditions to the output
	size_t tpi = 0;
	if (timepoints[tpi] == 0.0) {
		for (size_t i = 0; i < N; i++) {
			output(i, tpi) = NV_Ith_S(((N_Vector)y), i);
		}
		tpi++;
	}

#if 1
	// Simulate the system one step at a time
	size_t step_count = 0;
	while (tpi < (size_t)timepoints.size()) {
		// Make sure we don't take too many steps
		if (step_count >= MAX_CVODE_STEPS) {
			//LOGERROR("Maximum number of steps reached, stopping - %s.");
			failed_step_counts[MAX_CVODE_STEPS]++;
			return false;
		}
		step_count++;

		// Take one CVode step
		OdeReal tret;
		int result = CVode(cvode_mem, (OdeReal)timepoints[tpi], (N_Vector)y, &tret, CV_ONE_STEP);
		if (result != CV_SUCCESS) {
#if 0
			LOGERROR("CVode failed, last y values:");
			for (size_t i = 0; i < N; i++) {
				LOGERROR( "%u - %g", i, NV_Ith_S(y, i));
			}
#endif
			for (size_t j = tpi; j < (size_t)timepoints.size(); j++) {
				for (size_t i = 0; i < N; i++) {
					output(i, tpi) = std::numeric_limits<Real>::quiet_NaN();
				}
			}
			failed_step_counts[step_count]++;
			return false;
		}

		// If we pass a timepoint, interpolate back to that timepoint
		while (tret > timepoints[tpi]) {
			result = CVodeGetDky(cvode_mem, (OdeReal)timepoints[tpi], 0, (N_Vector)interpolate_y);
			if (result != CV_SUCCESS) {
				return false;
			}
			
			for (size_t i = 0; i < N; i++) {
				output(i, tpi) = NV_Ith_S(((N_Vector)interpolate_y), i);
			}

			tpi++;
			if (tpi >= (size_t)timepoints.size()) {
				break;
			}
		}
	}
#else
	// Simulate the system until each time point
	Real prevtime = 0.0;
	for (; tpi < timepoints.size(); tpi++) {
		if (timepoints[tpi] <= prevtime) {
			LOGERROR("Time points are not monotonically increasing");
			return false;
		}

		Real tret;
		int result = CVode(cvode_mem, timepoints[tpi], (N_Vector)y, &tret, CV_NORMAL);
		if (result != CV_SUCCESS) {
#if 0
			LOGERROR("CVode failed, last species values:");
			for (size_t i = 0; i < N; i++) {
				LOGERROR( "%u - %g", i, NV_Ith_S(y, i));
			}
#endif
			for (size_t j = tpi; j < timepoints.size(); j++) {
				for (size_t i = 0; i < N; i++) {
					output(i, tpi) = std::numeric_limits<Real>::quiet_NaN();
				}
			}
			return false;
		}
		
		for (size_t i = 0; i < N; i++) {
			output(i, tpi) = NV_Ith_S(((N_Vector)y), i);
		}

		prevtime = timepoints[tpi];
	}
#endif

	step_counts[step_count]++;

	return true;
}

void CVODESolver::DumpStatistics(const char* filename)
{
	FILE* file = fopen(filename, "w");
	if (file) {
		for (size_t i = 0; i <= MAX_CVODE_STEPS; i++) {
			fprintf(file, "%zu\t%zu\n", i, step_counts[i]);
		}
		for (size_t i = 0; i <= MAX_CVODE_STEPS; i++) {
			fprintf(file, "%zu\t%zu\n", i, failed_step_counts[i]);
		}

		fclose(file);
	}
}

int CVODESolver::cvode_rhs_fn(OdeReal t, OdeReal* y, OdeReal* ydot)
{
	if (!derivative) {
		return -2;
	}

	bool result = derivative(t, y, ydot, user_data);
	if (result) {
		return 0;
	} else {
		return -1;
	}
}

int CVODESolver::cvode_jac_fn(OdeReal t, OdeReal* y, OdeReal* ydot, OdeReal** jac)
{
	if (!jacobian) {
		return -2;
	}

	bool result = jacobian(t, y, ydot, jac, user_data);
	if (result) {
		return 0;
	} else {
		return -1;
	}
}

int CVODESolver::cvode_solver_init()
{
	ASSERT(linear_solver);
	if (!linear_solver->Init(N)) {
		return -1;
	} else {
		return 0;
	}
}

int CVODESolver::cvode_solver_setup(void* cvode_mem, int convfail, OdeReal* ypred, int* jcurPtr)
{
	ASSERT(linear_solver);
	CVodeMem cv_mem = (CVodeMem)cvode_mem;
	return linear_solver->Setup(cvode_mem, convfail, ypred, jcurPtr);
}

int CVODESolver::cvode_solver_solve(void* cvode_mem, OdeReal* b)
{
	ASSERT(linear_solver);
	return linear_solver->Solve(cvode_mem, b);
}

void CVODESolver::cvode_solver_free()
{
	ASSERT(linear_solver);
}
