#include "Utils.h"
#include "NetCDFDataFile.h"
#include "OutputHandlerNetCDF.h"

#ifdef BCM_INCLUDE_NETCDF

namespace bcm {

OutputHandlerNetCDF::OutputHandlerNetCDF()
	: NumVariables(0)
	, sample_ix(0)
	, AdditionalOutput(NULL)
{
}

OutputHandlerNetCDF::~OutputHandlerNetCDF()
{
}

void OutputHandlerNetCDF::SetFile(const std::string& filename)
{
	Filename = filename;
}

void OutputHandlerNetCDF::SetAdditionalOutput(AdditionalOutputInterfaceNetCDF* additional_output)
{
	AdditionalOutput = additional_output;
}

bool OutputHandlerNetCDF::Initialize(size_t sample_count, std::vector<std::string>& variables)
{
	// Open file in replace mode to remove any existing output
	if (!NetCDFFile.Open(Filename, netCDF::NcFile::replace)) {
		LOGERROR("Cannot open output file");
		return false;
	}

	NumVariables = variables.size();
	std::vector<unsigned int> sampleix(sample_count);
	for (size_t i = 0; i < sample_count; i++) {
		sampleix[i] = (unsigned int)i+1;
	}
	bool result = true;

	// Allocate variables
	result &= NetCDFFile.CreateGroup("samples");
	result &= NetCDFFile.CreateDimension("samples", "sample_ix", sampleix);
	result &= NetCDFFile.CreateDimension("samples", "variables", variables);
	result &= NetCDFFile.CreateVariable("samples", "variable_values", "sample_ix", "variables");
	result &= NetCDFFile.CreateVariable("samples", "log_prior", "sample_ix");
	result &= NetCDFFile.CreateVariable("samples", "log_likelihood", "sample_ix");
	result &= NetCDFFile.CreateVariable("samples", "log_posterior", "sample_ix");
	result &= NetCDFFile.CreateVariable("samples", "weight", "sample_ix");

	if (AdditionalOutput) {
		result &= AdditionalOutput->SetupAdditionalOutput(&NetCDFFile, sample_count);
	}

	NetCDFFile.Close();

	sample_ix = 0;
	return result;
}

void OutputHandlerNetCDF::Close()
{
	NetCDFFile.Close();
}

void OutputHandlerNetCDF::HandleSample(const Real* values, Real lposterior, Real llh, Real lprior, Real weight, void*)
{
	if (!NetCDFFile.Open(Filename, netCDF::NcFile::write)) {
		LOGERROR("Cannot open output file");
		return;
	}
	
	NetCDFFile.PutValue("samples", "sample_ix", sample_ix, (unsigned int)sample_ix);
	for (size_t vi = 0; vi < NumVariables; vi++) {
		NetCDFFile.PutValue("samples", "variable_values", sample_ix, vi, values[vi]);
	}
	NetCDFFile.PutValue("samples", "log_prior", sample_ix, lprior);
	NetCDFFile.PutValue("samples", "log_likelihood", sample_ix, llh);
	NetCDFFile.PutValue("samples", "log_posterior", sample_ix, lposterior);
	NetCDFFile.PutValue("samples", "weight", sample_ix, weight);

	NetCDFFile.Close();

	sample_ix++;
}

void OutputHandlerNetCDF::HandleAdditionalOutput(const Real* values, void*)
{
	if (!NetCDFFile.Open(Filename, netCDF::NcFile::write)) {
		LOGERROR("Cannot open output file");
		return;
	}

	if (AdditionalOutput) {
		AdditionalOutput->HandleAdditionalOutput(&NetCDFFile, sample_ix-1, values);
	}

	NetCDFFile.Close();
}

boost::function<void (const Real* values, Real lposterior, Real llh, Real lprior, Real weight, void* user)> OutputHandlerNetCDF::GetSampleFunctionPointer()
{
	return boost::bind(&OutputHandlerNetCDF::HandleSample, this, _1, _2, _3, _4, _5, _6);
}

boost::function<void (const Real* values, void* user)> OutputHandlerNetCDF::GetAdditionalOutputFunctionPointer()
{
	return boost::bind(&OutputHandlerNetCDF::HandleAdditionalOutput, this, _1, _2);
}

}

#endif
