#include "Utils.h"
#include "Prior.h"
#include "PriorGMM.h"
#include "PriorIndependence.h"
#include "PriorKDE.h"
#include "PriorVineCopula.h"
#include "VariableSet.h"

#include <boost/property_tree/xml_parser.hpp>

namespace bcm {

Prior::Prior(VariableSet* varset)
	: varset(varset)
{
	LowerBounds.resize(varset->GetNumVariables(), -std::numeric_limits<Real>::infinity());
	UpperBounds.resize(varset->GetNumVariables(), std::numeric_limits<Real>::infinity());
}

Prior::~Prior()
{
}

Prior* Prior::Create(const std::string& filename, VariableSet* varset)
{
	boost::property_tree::ptree pt;
	try {
		boost::property_tree::read_xml(filename, pt);
	} catch (boost::property_tree::xml_parser_error &e) {
		LOGERROR("Error loading variable file: %s", e.what());
		return NULL;
	}

	try {
		boost::property_tree::ptree prior_node;
		if (pt.count("prior") > 0) {
			prior_node = pt.get_child("prior");
		} else if (pt.count("variableset") > 0) {
			prior_node = pt.get_child("variableset");
		} else {
			LOGERROR("Incorrect prior XML format");
			return NULL;
		}

		std::string type = prior_node.get<std::string>("<xmlattr>.type", "");

		Prior* prior = NULL;
		if (type.empty() || type == "independence") {
			prior = new PriorIndependence(varset);
		} else if (type == "GMM") {
			prior = new PriorGMM(varset);
		} else if (type == "KDE") {
			prior = new PriorKDE(varset);
		} else if (type == "VineCopula") {
			prior = new PriorVineCopula(varset);
		} else {
			LOGERROR("Unknown prior type \"%s\"", type.c_str());
			return NULL;
		}

		if (prior->LoadFromXML(prior_node)) {
			return prior;
		} else {
			delete prior;
			return NULL;
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing variable file: %s", e.what());
		return NULL;
	}
}

bool Prior::LoadTransformationsFromXML(const boost::property_tree::ptree& xml_node)
{
	try {
		transforms.resize(varset->GetNumVariables(), Transform_Invalid);
		transform_parameters.resize(varset->GetNumVariables());

		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "variable_transformation") {
				std::string name = var.second.get<std::string>("<xmlattr>.name");
				size_t ix = varset->GetVariableIndex(name);
				if (ix == std::numeric_limits<size_t>::max()) {
					LOGERROR("Transformation specified for unknown variable \"%s\"", name.c_str());
					return false;
				}

				std::string transformation = var.second.get<std::string>("<xmlattr>.transform");
				if (transformation == "none") {
					transforms[ix] = Transform_None;
				} else if (transformation == "log") {
					transforms[ix] = Transform_Log;
					LowerBounds[ix] = 0.0;
				} else if (transformation == "logit") {
					transforms[ix] = Transform_Logit;
					LowerBounds[ix] = 0.0;
					UpperBounds[ix] = 1.0;
				} else if (transformation == "logit_scale") {
					transforms[ix] = Transform_LogitScale;
					
					std::string stra = var.second.get<std::string>("<xmlattr>.a");
					std::string strb = var.second.get<std::string>("<xmlattr>.b");
					Real a = boost::lexical_cast<Real>(stra);
					Real b = boost::lexical_cast<Real>(strb);
					transform_parameters[ix] = VectorReal(2);
					transform_parameters[ix](0) = a;
					transform_parameters[ix](1) = b;
					LowerBounds[ix] = a;
					UpperBounds[ix] = b;
				}
			}
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing prior file: %s", e.what());
		return false;
	}

	return true;
}

bool Prior::ApplyTransformations(const Real* values, VectorReal& v) const
{
	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		switch(transforms[i]) {
		case Transform_Log:
			if (values[i] <= 0.0) {
				return false;
			}
			v(i) = log(values[i]);
			break;

		case Transform_Logit:
			if (values[i] <= 0.0 || values[i] >= 1.0) {
				return false;
			}
			v(i) = logit(values[i]);
			break;

		case Transform_LogitScale:
			{
				const Real a = transform_parameters[i](0);
				const Real b = transform_parameters[i](1);
				if (values[i] <= a || values[i] >= b) {
					return false;
				}
				v(i) = logit_scale(values[i], a, b);
			}
			break;

		default:
			v(i) = values[i];
			break;
		}
	}

	return true;
}

bool Prior::CorrectProbabilityForTransform(Real& logp, const Real* values, VectorReal& v) const
{
	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		switch(transforms[i]) {
		case Transform_Log:
			//logp -= log(values[i]);
			logp -= v(i); // log(values[i]) is already stored in v(i)
			break;

		case Transform_Logit:
			// d(logit(x))/dx = 1/(x-x*x) ==> log(d(logit(x))/dx) = -log(x-x*x)
			{
				Real x = values[i];
				logp -= log(x-x*x);
			}
			break;

		case Transform_LogitScale:
			{
				const Real a = transform_parameters[i](0);
				const Real b = transform_parameters[i](1);
				logp += log(dlogit_scale(values[i], a, b));
			}
			break;

		default:
			break;
		}
	}

	return true;
}

bool Prior::ApplyInverseTransformations(Real* v) const
{
	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		switch(transforms[i]) {
		case Transform_Log:
			v[i] = exp(v[i]);
			break;

		case Transform_Logit:
			v[i] = logistic(v[i]);
			break;

		case Transform_LogitScale:
			{
				const Real ey = exp(v[i]);
				const Real a = transform_parameters[i](0);
				const Real b = transform_parameters[i](1);
				v[i] = (a + b*ey) / (ey+1);
			}
			break;

		default:
			break;
		}
	}

	return true;
}

}
