#include "Utils.h"
#include "PriorMarginalGMM.h"
#include "VariableSet.h"

#include <boost/property_tree/xml_parser.hpp>

namespace bcm {

PriorMarginalGMM::PriorMarginalGMM(VariableSet* varset)
	: Prior(varset)
{
	MarginalTypes.resize(varset->GetNumVariables(), MT_Invalid);
	Transforms.resize(varset->GetNumVariables(), Transform_Invalid);
	Marginals_Regular.resize(varset->GetNumVariables(), NULL);
	Marginals_GMM.resize(varset->GetNumVariables(), NULL);
}

PriorMarginalGMM::~PriorMarginalGMM()
{
}

bool PriorMarginalGMM::LoadFromXML(const std::string& filename)
{
	boost::property_tree::ptree pt;
	try {
		boost::property_tree::read_xml(filename, pt);
	} catch (boost::property_tree::xml_parser_error &e) {
		LOGERROR("Error loading variable file: %s", e.what());
		return false;
	}
	
	try {
		boost::property_tree::ptree model_node = pt.get_child("variableset");
		
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, model_node.get_child("")) {
			if (var.first == "variable") {
				std::string name = var.second.get<std::string>("<xmlattr>.name");
				size_t ix = varset->GetVariableIndex(name);
				if (ix == std::numeric_limits<size_t>::max()) {
					LOGERROR("Unknown variable \"%s\"", name.c_str());
					return false;
				}

				boost::optional<std::string> type = var.second.get_optional<std::string>("<xmlattr>.type");
				if (type) {
					if (*type == "regular") {
						UnivariateMarginal* m = new UnivariateMarginal;
						if (!m->Initialize(var.second)) {
							return false;
						}
						Marginals_Regular[ix] = m;
						Transforms[ix] = Transform_None;
						LowerBounds[ix] = m->GetLowerBound();
						UpperBounds[ix] = m->GetUpperBound();
						MarginalTypes[ix] = MT_Regular;
					} else if (*type == "gmm" || *type == "gmm_log" || *type == "gmm_logit") {
						UnivariateMarginalGMM* m = new UnivariateMarginalGMM;
						if (!m->Initialize(var.second)) {
							return false;
						}
						Marginals_GMM[ix] = m;
						MarginalTypes[ix] = MT_GMM;

						if (*type == "gmm_log") {
							Transforms[ix] = Transform_Log;
							LowerBounds[ix] = 0.0;
							UpperBounds[ix] = std::numeric_limits<Real>::infinity();
						} else if (*type == "gmm_logit") {
							Transforms[ix] = Transform_Logit;
							LowerBounds[ix] = 0.0;
							UpperBounds[ix] = 1.0;
						} else {
							Transforms[ix] = Transform_None;
							LowerBounds[ix] = -std::numeric_limits<Real>::infinity();
							UpperBounds[ix] = std::numeric_limits<Real>::infinity();
						}
					} else {
						LOGERROR("Invalid type \"%s\"", type->c_str());
						return false;
					}
				} else {
					UnivariateMarginal* m = new UnivariateMarginal;
					if (!m->Initialize(var.second)) {
						return false;
					}
					Marginals_Regular[ix] = m;
					Transforms[ix] = Transform_None;
					LowerBounds[ix] = m->GetLowerBound();
					UpperBounds[ix] = m->GetUpperBound();
					MarginalTypes[ix] = MT_Regular;
				}

				boost::optional<std::string> transform = var.second.get_optional<std::string>("<xmlattr>.transform");
				if (transform) {
					if (*transform == "log") {
						Transforms[ix] = Transform_Log;
						LowerBounds[ix] = 0.0;
						UpperBounds[ix] = std::numeric_limits<Real>::infinity();
					} else if (*transform == "logit") {
						Transforms[ix] = Transform_Logit;
						LowerBounds[ix] = 0.0;
						UpperBounds[ix] = 1.0;
					} else if (*transform == "none") {
						Transforms[ix] = Transform_None;
					} else {
						LOGERROR("Invalid transform \"%s\"", transform->c_str());
						return false;
					}
				}
			}
		}

		return true;
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing variable file: %s", e.what());
		return false;
	}
}

bool PriorMarginalGMM::EvaluateLogPDF(size_t threadix, const Real* values, Real& logp) const
{
	logp = 0.0;

	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		Real varlogp;
		if (!EvaluateMarginalLogPDF(i, values[i], varlogp)) {
			return false;
		}
		logp += varlogp;
	}

	return true;
}

bool PriorMarginalGMM::Sample(Real* values, RNG* rng) const
{
	for (size_t i = 0; i < varset->GetNumVariables(); i++) {
		Real x = std::numeric_limits<Real>::quiet_NaN();

		if (MarginalTypes[i] == MT_Regular) {
			if (!Marginals_Regular[i]->Sample(x, rng)) {
				return false;
			}
		} else if (MarginalTypes[i] == MT_GMM) {
			if (!Marginals_GMM[i]->Sample(x, rng)) {
				return false;
			}
		} else {
			LOGERROR("Unknown marginal type for variable %u", i);
			return false;
		}
		
		if (Transforms[i] == Transform_Log) {
			x = exp(x);
		} else if(Transforms[i] == Transform_Logit) {
			x = 1.0 / (1.0 + exp(-x));
		}

		values[i] = x;
	}

	return true;
}

bool PriorMarginalGMM::EvaluateMarginalVariance(size_t i, Real& var) const
{
	if (i >= varset->GetNumVariables()) {
		return false;
	}

	if (Transforms[i] == Transform_Logit) {
		// No formula for variance, but at least it's not going to be more than 1/12 (variance of uniform(0,1))
		var = 1.0 / 24.0;
		return true;
	}

	if (MarginalTypes[i] == MT_Regular) {
		return Marginals_Regular[i]->EvaluateVariance(var);
	} else if (MarginalTypes[i] == MT_GMM) {
		if (Transforms[i] == Transform_Log) {
			return Marginals_GMM[i]->EvaluateVarianceLogTransformed(var);
		} else {
			return Marginals_GMM[i]->EvaluateVariance(var);
		}
	} else {
		LOGERROR("Unknown marginal type for variable %u", i);
		return false;
	}
}

bool PriorMarginalGMM::EvaluateMarginalLogPDF(size_t i, Real value, Real& logp) const
{
	Real x;
	if (Transforms[i] == Transform_Log) {
		if (value <= 0.0) {
			logp = -std::numeric_limits<Real>::infinity();
			return true;
		}
		x = log(value);
	} else if(Transforms[i] == Transform_Logit) {
		if (value <= 0.0 || value >= 1.0) {
			logp = -std::numeric_limits<Real>::infinity();
			return true;
		}
		x = log(value / (1.0 - value));
	} else {
		x = value;
	}

	if (MarginalTypes[i] == MT_Regular) {
		if (!Marginals_Regular[i]->EvaluateLogPDF(logp, x)) {
			return false;
		}
	} else if (MarginalTypes[i] == MT_GMM) {
		if (!Marginals_GMM[i]->EvaluateLogPDF(logp, x)) {
			return false;
		}
	} else {
		LOGERROR("Unknown marginal type for variable %u", i);
		return false;
	}
	
	// Transformed PDF = PDF of transform times the absolute value of the derivative of the transformation
	if (Transforms[i] == Transform_Log) {
		// Don't need to recalculate log(value), it's already stored in x
		logp -= x;
	} else if (Transforms[i] == Transform_Logit) {
		logp -= log(value * (1.0 - value));
	}

	return true;
}

}
