#pragma once

#include "SamplerSMC.h"

namespace bcm {

class SMCSamplingStrategy
{
public:
	SMCSamplingStrategy(SamplerSMC* sampler, TaskManager* task_manager);
	virtual ~SMCSamplingStrategy();

	//! Update the sampling strategy; this is called once per sequential sampling step, before UpdateParticle.
	//! This will call UpdateStrategyImpl, which can be overridden by derived classes.
	bool UpdateStrategyIteration(bool force_update, RNG* rng);
	bool UpdateStrategyCurrent(RNG* rng);
	
	//! Update one particle.
	bool UpdateParticle(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix);

	//! Calculate weight of a new particle
	bool CalculateParticleWeight(SamplerSMC::Particle& particle, size_t ix, size_t threadix);

protected:
	//! Virtual function to be overridden by derived class, which is called by UpdateStrategy.
	virtual bool UpdateStrategyIterationImpl(bool force_update, RNG* rng) { return true; }
	virtual bool UpdateStrategyCurrentImpl(RNG* rng) { return true; }
	
	//! Pure virtual function to be overridden by derived class, which is called by UpdateParticle.
	//! You're given a random number generator, and an index of which particle this is in case you want to store information over the course of multiple
	//! sampling rounds. The index ceases to be informative after a resampling step.
	//! This function is expected to be re-entrant, so store information only using the stack or the provided index. The rng can be safely used as there as a unique
	//! rng per thread.
	virtual bool UpdateParticleImpl(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix) = 0;
	virtual bool CalculateParticleWeightImpl(SamplerSMC::Particle& particle, size_t ix, size_t threadix) = 0;

	bool EvaluatePriorLikelihood(size_t threadix, const Real* values, Real& lprior, Real& llh);
	bool GetOutputValues(size_t threadix, Real* values);
	VectorReal CalculatePreviousPopulationMean();
	VectorReal CalculatePreviousPopulationVariance(const VectorReal& mean);
	MatrixReal CalculatePreviousPopulationCovariance(const VectorReal& mean);

	SamplerSMC* sampler;
	const Prior* prior;
	TaskManager* task_manager;
	size_t num_variables;
	size_t population_size;
	size_t num_threads;

	const SamplerSMC::TParticleVector* previous_population;
	Real current_temperature;
	Real previous_temperature;
};

}
