#include "Utils.h"
#include "SamplerSMC.h"
#include "SMCSamplingStrategyKDE.h"

namespace bcm {

SMCSamplingStrategyKDE::SMCSamplingStrategyKDE(SamplerSMC* sampler, TaskManager* task_manager)
	: SMCSamplingStrategy(sampler, task_manager)
	, bandwidth_method(BM_DiagonalSilverman)
	, empirical_mean(VectorReal::Zero(num_variables))
	, empirical_variance(VectorReal::Zero(num_variables))
	, diagonal_bandwidth(VectorReal::Zero(num_variables))
	, one_over_two_bw_sq(VectorReal::Zero(num_variables))
	, log_normalizing_constant(0.0)
{
}

SMCSamplingStrategyKDE::~SMCSamplingStrategyKDE()
{
}

bool SMCSamplingStrategyKDE::UpdateStrategyIterationImpl(bool force_update, RNG* rng)
{
	if (bandwidth_method == BM_DiagonalSilverman) {
		// Calculate empirical mean
		empirical_mean.setZero();
		for (size_t k = 0; k < previous_population->size(); k++) {
			const VectorReal& s = (*previous_population)[k].var_values;
			Real one_over_kp1 = 1.0 / (k + 1.0);
			for (size_t i = 0; i < num_variables; i++) {
				empirical_mean(i) += (s(i) - empirical_mean(i)) * one_over_kp1;
			}
		}

		// Calculate empirical variance
		empirical_variance.setZero();
		for (size_t k = 0; k < previous_population->size(); k++) {
			const VectorReal& s = (*previous_population)[k].var_values;
			Real one_over_kp1 = 1.0 / (k + 1.0);
			for (size_t i = 0; i < num_variables; i++) {
				Real d = s(i) - empirical_mean(i);
				empirical_variance(i) += (d * d - empirical_variance(i)) * one_over_kp1;
			}
		}
		Real n = (Real)previous_population->size();
		empirical_variance *= n / (n - 1.0);
	
		// Set diagonal bandwidth for the next iteration
		log_normalizing_constant = 0.0;
		for (size_t i = 0; i < num_variables; i++) {
			Real var = empirical_variance(i);
			Real bw = (Real)1.05922 * sqrt(var) * pow((Real)population_size, (Real)-0.2);
			diagonal_bandwidth(i) = bw;
			one_over_two_bw_sq(i) = (Real)1.0 / ((Real)2.0 * bw * bw);
			log_normalizing_constant += log(rsqrt((Real)2.0 * M_PI * bw * bw));
		}
	} else {
		LOGERROR("Invalid bandwidth method");
		return false;
	}

	return true;
}

bool SMCSamplingStrategyKDE::UpdateParticleImpl(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix)
{
	// Propose new parameters
	if (bandwidth_method == BM_DiagonalSilverman) {
		for (size_t i = 0; i < num_variables; i++) {
			Real x = rng->GetNormal(0.0, diagonal_bandwidth(i));
			particle.var_values[i] += x;
		}
	} else {
		LOGERROR("Invalid proposal type");
		return false;
	}

	// Evaluate prior & likelihood
	bool result = EvaluatePriorLikelihood(threadix, particle.var_values.data(), particle.lprior, particle.llh);
	if (!result) {
		return false;
	}
	particle.lpowerposterior = particle.lprior + current_temperature * particle.llh;
	if (particle.output_values.size() > 0) {
		if (!GetOutputValues(threadix, particle.output_values.data())) {
			return false;
		}
	}

	return true;
}

bool SMCSamplingStrategyKDE::CalculateParticleWeightImpl(SamplerSMC::Particle& particle, size_t ix, size_t threadix)
{
	// Calculate new weight
	if (particle.lpowerposterior != -std::numeric_limits<Real>::infinity()) {
		Real lproposal = -std::numeric_limits<Real>::infinity();
		for (size_t j = 0; j < population_size; j++) {
			const SamplerSMC::Particle& prevj = (*previous_population)[j];
		
			Real thislp = 0.0;
			for (size_t i = 0; i < num_variables; i++) {
				Real d = particle.var_values(i) - prevj.var_values(i);
				thislp -= d * d * one_over_two_bw_sq(i);
			}
			thislp += log_normalizing_constant;
			lproposal = logsum(lproposal, thislp);
		}

		Real lweight = particle.lpowerposterior - lproposal;
		particle.lweight = lweight - particle.lnormweight;
		particle.accumulated_lweight = particle.lweight;
	} else {
		particle.lweight = -std::numeric_limits<Real>::infinity();
		particle.accumulated_lweight = -std::numeric_limits<Real>::infinity();
	}

	return true;
}

}
