#include "Utils.h"
#include "Prior.h"
#include "SamplerSMC.h"
#include "SMCSamplingStrategyMCMC.h"

namespace bcm {

SMCSamplingStrategyMCMC::SMCSamplingStrategyMCMC(SamplerSMC* sampler, TaskManager* task_manager)
	: SMCSamplingStrategy(sampler, task_manager)
	, proposal_type(PT_GlobalEmpiricalCovariance)
	, num_nearby_particles(50)
	, recalculate_steps(20)
	, num_mcmc_steps_per_iteration(1)
	, empirical_mean(VectorReal::Zero(num_variables))
	, empirical_variance(VectorReal::Zero(num_variables))
	, proposal_sigma(VectorReal::Zero(num_variables))
	, empirical_covariance(MatrixReal::Zero(num_variables, num_variables))
	, update_step(0)
	, target_acceptance_rate(0.23)
	, proposal_scale(2.4 * rsqrt((Real)num_variables))
{
	async.resize(num_threads);
	for (size_t k = 0; k < num_threads; k++) {
		async[k].proposed_var_values.resize(num_variables);
		async[k].num_accepted = 0;
		async[k].num_proposed = 0;
	}
}

SMCSamplingStrategyMCMC::~SMCSamplingStrategyMCMC()
{
}

void SMCSamplingStrategyMCMC::SetGlobal()
{
	proposal_type = PT_GlobalEmpiricalVariance;
}

void SMCSamplingStrategyMCMC::SetGlobalCovariance()
{
	proposal_type = PT_GlobalEmpiricalCovariance;
}

void SMCSamplingStrategyMCMC::SetLocal(size_t num_nearby)
{
	ASSERT(num_nearby > 2);
	proposal_type = PT_LocalEmpiricalVariance;
	num_nearby_particles = num_nearby;
	local_proposals.resize(population_size);
	for (size_t k = 0; k < population_size; k++) {
		local_proposals[k].proposal_sigma = VectorReal::Zero(num_variables);
	}
}

void SMCSamplingStrategyMCMC::SetStepsPerIteration(size_t n)
{
	ASSERT(n > 0);
	num_mcmc_steps_per_iteration = n;
}

void SMCSamplingStrategyMCMC::SetRecalculateSteps(size_t n)
{
	ASSERT(n > 0);
	recalculate_steps = n;
}

bool SMCSamplingStrategyMCMC::UpdateStrategyIterationImpl(bool force_update, RNG* rng)
{
	if (force_update || update_step == recalculate_steps) {
		if (proposal_type == PT_GlobalEmpiricalVariance) {
			empirical_mean = CalculatePreviousPopulationMean();
			empirical_variance = CalculatePreviousPopulationVariance(empirical_mean);
	
			// Set proposal variance for the next iteration
			for (size_t i = 0; i < num_variables; i++) {
				Real prior_var;
				if (!prior->EvaluateMarginalVariance(i, prior_var)) {
					// TODO - somehow come up with something reasonable?
					prior_var = 1.0;
				}
				Real proposal_var = std::max(empirical_variance(i), (Real)1e-6 * prior_var);
				proposal_sigma(i) = sqrt(proposal_var);
			}
		} else if (proposal_type == PT_GlobalEmpiricalCovariance) {
			empirical_mean = CalculatePreviousPopulationMean();

			// Calculate empirical covariance
			empirical_covariance.setZero();
			Real acc_weight = 0;
			for (size_t k = 0; k < previous_population->size(); k++) {
				const VectorReal& s = (*previous_population)[k].var_values;
				const Real w = (*previous_population)[k].normweight * previous_population->size();
				Real one_over_kp1 = w / (acc_weight + w);
				for (size_t i = 0; i < num_variables; i++) {
					for (size_t j = 0; j < num_variables; j++) {
						const Real d1 = s(i) - empirical_mean(i);
						const Real d2 = s(j) - empirical_mean(j);
						empirical_covariance(i,j) += (d1 * d2 - empirical_covariance(i,j)) * one_over_kp1;
					}
				}
				acc_weight += w;
			}
			Real n = (Real)previous_population->size();
			empirical_covariance *= n / (n - 1.0);
			
			// Set proposal covariance for the next iteration
			mvn_sampler.SetCovariance(empirical_covariance);
		} else if (proposal_type == PT_LocalEmpiricalVariance) {
			std::vector<uint64> tasks;
			tasks.resize(task_manager->GetNumThreads());
			for (size_t i = 0; i < tasks.size(); i++) {
				TTask task = boost::bind(&SMCSamplingStrategyMCMC::UpdateStrategyLocalMCMC, this, _1, _2);
				tasks[i] = task_manager->AddTask(task, reinterpret_cast<void*>(i));
			}
			for (size_t i = 0; i < tasks.size(); i++) {
				task_manager->WaitTask(tasks[i]);
			}
		} else {
			LOGERROR("Invalid proposal type");
			return false;
		}

		update_step = 1;
	} else {
		update_step++;
	}

	return true;
}

bool SMCSamplingStrategyMCMC::UpdateStrategyCurrentImpl(RNG* rng)
{
	size_t num_accepted = 0;
	size_t num_proposed = 0;
	for (size_t k = 0; k < num_threads; k++) {
		num_accepted += async[k].num_accepted;
		num_proposed += async[k].num_proposed;
		async[k].num_accepted = 0;
		async[k].num_proposed = 0;
	}

	Real acceptance_rate = (Real)num_accepted / (Real)num_proposed;
	//LOG("Acceptance rate: %g - scale: %g", acceptance_rate, proposal_scale);
	
	// Adjust scaling to try to get the target acceptance rate.
	Real learn_rate = 1.0 + rng->GetReal() * 0.1;
	if (acceptance_rate < 0.91 * target_acceptance_rate) {
		proposal_scale /= learn_rate;
		proposal_scale = std::max(proposal_scale, (Real)1e-10);
	} else if (acceptance_rate > 1.1 * target_acceptance_rate) {
		proposal_scale *= learn_rate;
		proposal_scale = std::min(proposal_scale, (Real)1e3);
	}

	return true;
}

bool SMCSamplingStrategyMCMC::UpdateParticleImpl(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix)
{
	bool result;

	for (size_t step = 0; step < num_mcmc_steps_per_iteration; step++) {
#if 1
		// Propose new parameters
		if (proposal_type == PT_GlobalEmpiricalVariance) {
			for (size_t i = 0; i < num_variables; i++) {
				Real x = rng->GetNormal(0.0, proposal_scale * proposal_sigma(i));
				Real nx = particle.var_values[i] + x;
				async[threadix].proposed_var_values(i) = nx;
			}
		} else if (proposal_type == PT_GlobalEmpiricalCovariance) {
			VectorReal& x = async[threadix].proposed_var_values;
			mvn_sampler.Get(rng, x);
			x *= proposal_scale;
			x += particle.var_values;
		} else if (proposal_type == PT_LocalEmpiricalVariance) {
			LocalProposal& proposal = local_proposals[ix];
			for (size_t i = 0; i < num_variables; i++) {
				Real x = rng->GetNormal(0.0, proposal_scale * proposal.proposal_sigma(i));
				Real nx = particle.var_values[i] + x;
				async[threadix].proposed_var_values(i) = nx;
			}
		} else {
			LOGERROR("Invalid proposal type");
			return false;
		}
		
		// Reflect for variables with bounded support, this is still a symmetrical proposal
		for (size_t i = 0; i < num_variables; i++) {
			const Real lb = prior->GetLowerBound(i);
			const Real ub = prior->GetUpperBound(i);
			Real nx = async[threadix].proposed_var_values(i);
			while (1) {
				if (nx < lb) {
					nx = lb + (lb - nx);
				} else if (nx > ub) {
					nx = ub - (nx - ub);
				} else {
					break;
				}
			}
			async[threadix].proposed_var_values(i) = nx;
		}

		// Evaluate prior & likelihood
		Real new_lprior;
		Real new_llh;
		Real new_lpowerposterior;
		result = EvaluatePriorLikelihood(threadix, async[threadix].proposed_var_values.data(), new_lprior, new_llh);
		if (!result) {
			return false;
		}
		new_lpowerposterior = new_lprior + current_temperature * new_llh;

		// Decide on transition
		Real transition_probability = new_lpowerposterior - particle.lpowerposterior;
		transition_probability = exp(transition_probability);
		transition_probability = std::min(transition_probability, (Real)1.0);

		Real alpha = rng->GetReal();
		async[threadix].num_proposed++;
		if (alpha < transition_probability) {
			async[threadix].num_accepted++;

			particle.var_values = async[threadix].proposed_var_values;
			if (particle.output_values.size() > 0) {
				if (!GetOutputValues(threadix, particle.output_values.data())) {
					return false;
				}
			}
			particle.lprior = new_lprior;
			particle.llh = new_llh;
			particle.lpowerposterior = new_lpowerposterior;
		} else {
			// Revert to previous state
		}
#else
		for (size_t i = 0; i < num_variables; i++) {
			// Propose new parameters
			async[threadix].proposed_var_values = particle.var_values;
			if (proposal_type == PT_GlobalEmpiricalVariance) {
				Real x = rng->GetNormal(0.0, proposal_scale * proposal_sigma(i));
				Real nx = particle.var_values[i] + x;
				async[threadix].proposed_var_values(i) = nx;

				if (nx != nx) {
					LOGERROR("Invalid value %g, prev %g, var %g, scale %g", nx, x, proposal_sigma(i), proposal_scale);
				}
			} else {
				LOGERROR("Invalid proposal type");
				return false;
			}
		
			// Reflect for variables with bounded support, this is still a symmetrical proposal
			const Real lb = prior->GetLowerBound(i);
			const Real ub = prior->GetUpperBound(i);
			Real nx = async[threadix].proposed_var_values(i);
			while (1) {
				if (nx < lb) {
					nx = lb + (lb - nx);
				} else if (nx > ub) {
					nx = ub - (nx - ub);
				} else {
					break;
				}
			}
			async[threadix].proposed_var_values(i) = nx;

			// Evaluate prior & likelihood
			Real new_lprior;
			Real new_llh;
			Real new_lpowerposterior;
			result = EvaluatePriorLikelihood(threadix, async[threadix].proposed_var_values.data(), new_lprior, new_llh);
			if (!result) {
				return false;
			}
			new_lpowerposterior = new_lprior + current_temperature * new_llh;

			// Decide on transition
			Real transition_probability = new_lpowerposterior - particle.lpowerposterior;
			transition_probability = exp(transition_probability);
			transition_probability = std::min(transition_probability, (Real)1.0);

			Real alpha = rng->GetReal();
			async[threadix].num_proposed++;
			if (alpha < transition_probability) {
				async[threadix].num_accepted++;

				particle.var_values = async[threadix].proposed_var_values;
				if (particle.output_values.size() > 0) {
					if (!GetOutputValues(threadix, particle.output_values.data())) {
						return false;
					}
				}
				particle.lprior = new_lprior;
				particle.llh = new_llh;
				particle.lpowerposterior = new_lpowerposterior;
			} else {
				// Revert to previous state
			}
		}
#endif
	}
	
	return true;
}

bool SMCSamplingStrategyMCMC::CalculateParticleWeightImpl(SamplerSMC::Particle& particle, size_t ix, size_t threadix)
{
	const SamplerSMC::Particle& prev_particle = (*previous_population)[ix];

	// Calculate new weight with approximate backward kernel
	if (prev_particle.normweight > 0) {
		Real incremental_weight = (prev_particle.lprior + current_temperature * prev_particle.llh) - prev_particle.lpowerposterior;
		particle.lweight = incremental_weight;
		particle.accumulated_lweight = prev_particle.accumulated_lweight + incremental_weight;
	} else {
		particle.lweight = -std::numeric_limits<Real>::infinity();
		particle.accumulated_lweight = -std::numeric_limits<Real>::infinity();
	}

	return true;
}

void SMCSamplingStrategyMCMC::UpdateStrategyLocalMCMC(void* user, size_t thread)
{
	size_t ti = reinterpret_cast<size_t>(user);
	size_t nt = task_manager->GetNumThreads();
	size_t np = previous_population->size();

	size_t start_particle = (ti * np) / nt;
	size_t end_particle = ((ti + 1) * np) / nt;
	end_particle = std::min(end_particle, np);
	
	std::vector<size_t> nearby_particles;
	std::vector<Real> nearby_particles_distSq;

	nearby_particles.reserve(num_nearby_particles);
	nearby_particles_distSq.reserve(num_nearby_particles);
	Real largest_distSq;

	// For each particle, calculate the local variance by taking those other particles
	// which are close in the other dimensions
	for (size_t k = start_particle; k < end_particle; k++) {
		const VectorReal& s1 = (*previous_population)[k].var_values;
		LocalProposal& proposal = local_proposals[k];
				
		for (size_t i = 0; i < num_variables; i++) {
			nearby_particles.clear();
			nearby_particles_distSq.clear();
			largest_distSq = 0;

			for (size_t l = 0; l < previous_population->size(); l++) {
				if (l == k) {
					continue;
				}
						
				// Calculate distance project on current variable
				const VectorReal& s2 = (*previous_population)[l].var_values;
				Real distSq = 0.0;
				for (size_t j = 0; j < num_variables; j++) {
					if (j == i) {
						continue;
					}
					const Real d = s2(j) - s1(j);
					distSq += d*d;
				}

				// If this particle is closer than the furthest particle in the set, add
				// the particle to the set of closest particle and remove the furthest.
				if (nearby_particles.size() < num_nearby_particles) {
					nearby_particles.push_back(l);
					nearby_particles_distSq.push_back(distSq);
					if (distSq > largest_distSq) {
						largest_distSq = distSq;
					}
				} else if (distSq < largest_distSq) {
					Real largest = 0;
					size_t largest_ix = 0;
					Real second_largest = 0;
					for (size_t m = 0; m < nearby_particles.size(); m++) {
						if (nearby_particles_distSq[m] > largest) {
							if (largest > second_largest) {
								second_largest = largest;
							}
							largest = nearby_particles_distSq[m];
							largest_ix = m;
						} else if(nearby_particles_distSq[m] > second_largest) {
							second_largest = nearby_particles_distSq[m];
						}
					}

					nearby_particles[largest_ix] = l;
					nearby_particles_distSq[largest_ix] = distSq;
					largest_distSq = std::max(distSq, second_largest);
				}
			}

			// Given the set of close particles, calculate local variance for this variable
			// Calculate empirical mean
			Real empirical_mean = (*previous_population)[k].var_values(i);
			for (size_t j = 0; j < nearby_particles.size(); j++) {
				const size_t ix = nearby_particles[j];
				const Real s = (*previous_population)[ix].var_values(i);
				empirical_mean += (s - empirical_mean) / (j + 2.0); // +2 because j starts at 0 whereas it should be at 1
			}

			// Calculate empirical variance
			Real d = (*previous_population)[k].var_values(i) - empirical_mean;
			Real empirical_variance = d * d;
			for (size_t j = 0; j < nearby_particles.size(); j++) {
				const size_t ix = nearby_particles[j];
				const Real d = (*previous_population)[ix].var_values(i) - empirical_mean;
				empirical_variance += (d * d - empirical_variance) / (j + 2.0); // +2 because j starts at 0 whereas it should be at 1
			}
			Real n = (Real)nearby_particles.size() + 1;
			empirical_variance *= n / (n - 1.0);
	
			// Set proposal variance for the next iteration
			proposal.proposal_sigma(i) = sqrt(empirical_variance);
		}
	}
}

}
