#include "Utils.h"
#include "RNG.h"
#include "Sample.h"
#include "SamplePopulation.h"

namespace bcm {

SamplePopulation::SamplePopulation()
	: num_variables(0)
	, num_output(0)
{
}

SamplePopulation::~SamplePopulation()
{
}

void SamplePopulation::Reserve(size_t n_samples, size_t n_variables, size_t n_output)
{
	num_variables = n_variables;
	num_output = n_output;

	free_sample_pool.resize(n_samples);
	for (size_t i = 0; i < n_samples; i++) {
		Sample* s = new Sample(n_variables, n_output);
		free_sample_pool[i] = s;
	}
}

std::list<Sample*>::iterator SamplePopulation::GetLowestLLHSample()
{
	if (samples.empty()) {
		return samples.end();
	}

	std::list<Sample*>::iterator min_sample = samples.begin();
	Real min_llh = std::numeric_limits<Real>::infinity();
	for (std::list<Sample*>::iterator si = samples.begin(); si != samples.end(); ++si) {
		if ((*si)->llh < min_llh) {
			min_llh = (*si)->llh;
			min_sample = si;
		}
	}

	return min_sample;
}

std::list<Sample*>::iterator SamplePopulation::GetHighestLLHSample()
{
	if (samples.empty()) {
		return samples.end();
	}

	std::list<Sample*>::iterator max_sample = samples.begin();
	Real max_llh = -std::numeric_limits<Real>::infinity();
	for (std::list<Sample*>::iterator si = samples.begin(); si != samples.end(); ++si) {
		if ((*si)->llh > max_llh) {
			max_llh = (*si)->llh;
			max_sample = si;
		}
	}

	return max_sample;
}

std::list<Sample*>::iterator SamplePopulation::GetRandomSample(RNG* rng)
{
	unsigned long ix = rng->GetUniformInt(static_cast<unsigned long>(samples.size()));
	std::list<Sample*>::iterator si = samples.begin();
	for (unsigned long i = 0; i < ix; i++) {
		++si;
	}
	return si;
}

std::list<Sample*>::const_iterator SamplePopulation::GetRandomSample(RNG* rng) const
{
	unsigned long ix = rng->GetUniformInt(static_cast<unsigned long>(samples.size()));
	std::list<Sample*>::const_iterator si = samples.cbegin();
	for (unsigned long i = 0; i < ix; i++) {
		++si;
	}
	return si;
}

void SamplePopulation::RemoveSample(std::list<Sample*>::iterator sample)
{
	free_sample_pool.push_back(*sample);
	samples.erase(sample);
}

Sample* SamplePopulation::AllocateSample()
{
	Sample* new_sample = NULL;
	if (free_sample_pool.empty()) {
		new_sample = new Sample(num_variables, num_output);
	} else {
		new_sample = free_sample_pool.back();
		free_sample_pool.pop_back();
	}
	return new_sample;
}

void SamplePopulation::AddSample(Sample* new_sample)
{
	samples.push_back(new_sample);
}

void SamplePopulation::FreeSample(Sample* new_sample)
{
	free_sample_pool.push_back(new_sample);
}

std::list<Sample*>::const_iterator SamplePopulation::cbegin() const
{
	return samples.cbegin();
}

std::list<Sample*>::const_iterator SamplePopulation::cend() const
{
	return samples.end();
}

size_t SamplePopulation::size() const
{
	return samples.size();
}

}
