#include "Utils.h"
#include "SamplerFactory.h"
#include "SamplerIS.h"
#include "SamplerMH.h"
#include "SamplerNested.h"
#include "SamplerPT.h"
#include "SamplerSMC.h"

namespace po = boost::program_options;

namespace bcm {

Sampler* SamplerFactory::Create(const char* name, size_t numthreads)
{
	Sampler* sampler = NULL;

	if (strcmp(name, "mh") == 0 || strcmp(name, "metropolis_hastings") == 0) {
		sampler = new SamplerMH(numthreads);
	} else if (strcmp(name, "nested") == 0) {
		sampler = new SamplerNested(numthreads);
	} else if (strcmp(name, "pt") == 0 || strcmp(name, "parallel_tempering") == 0) {
		sampler = new SamplerPT(numthreads);
	} else if (strcmp(name, "smc") == 0 || strcmp(name, "sequential_monte_carlo") == 0) {
		sampler = new SamplerSMC(numthreads);
	} else if (strcmp(name, "is") == 0) {
		sampler = new SamplerIS(numthreads);
	}

	return sampler;
}

Sampler* SamplerFactory::Create(const boost::program_options::variables_map& vm, size_t numthreads)
{
	Sampler* sampler = NULL;

	std::string type;
	try {
		type = vm["sampler.type"].as<std::string>();
	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler config: %s", e.what());
		return NULL;
	}

	sampler = Create(type.c_str(), numthreads);
	if (!sampler) {
		return NULL;
	}

	if (sampler->Parse(vm)) {
		return sampler;
	} else {
		delete sampler;
		return NULL;
	}
}

void SamplerFactory::AddOptionsDescription(po::options_description& pod)
{
	pod.add_options()
		("sampler.type",								po::value<std::string>()->default_value(""),					"Sampler type, can be \"mh\", \"pt\", \"smc\" or \"nested\".")
		("sampler.burnin_period",						po::value<size_t>()->default_value(1000),						"[MH/PT] Number of samples for burnin-period, these are discarded.")
		("sampler.sampling_period",						po::value<size_t>()->default_value(5000),						"[MH/PT] Number of samples to produce, after the burnin period.")
		("sampler.use_every_nth",						po::value<size_t>()->default_value(1),							"[MH/PT] Subsampling, a value of 10 means that every 10th sample is used and the 9 samples in between are discarded.")
		("sampler.num_mcmc_moves_per_iteration",		po::value<size_t>()->default_value(1),							"[MH/PT] Number of mcmc moves to make per iteration, mainly useful for decreasing the number of context switches when using multiple threads.")
		("sampler.proposal_scaling",					po::value<Real>()->default_value(1.0),							"[MH] Scaling factor for the proposal distribution.")
		("sampler.numchains",							po::value<size_t>()->default_value(8),							"[PT] Number of chains")
		("sampler.min_temperature",						po::value<Real>()->default_value(0.0),							"[PT] Temperature of the lowest chain. Note that the evidence calculation is not correct if this is not set to 0.")
		("sampler.adapt_proposal",						po::value<size_t>()->default_value(2500),						"[PT] Number of samples after which the proposal variance should be adapted, 0 for no adaptation.")
		("sampler.exchange_probability",				po::value<Real>()->default_value(0.5),							"[PT] Probability of choosing an exchange step instead of a MH-move.")
		("sampler.temperatures",						po::value<std::string>()->default_value(""),					"[PT] Filename specifying a file containing the temperatures to be used for each chain, if empty use the default temperature schedule.")
		("sampler.initial_temperature_power",			po::value<Real>()->default_value(5.0),							"[PT] Specifies the rate of the power-law used for the initial temperature schedule, used if min_temperature=0.")
		("sampler.adapt_temperatures_times",			po::value<size_t>()->default_value(0),							"[PT] Number of iterations of temperature adaptation steps.")
		("sampler.adapt_temperatures_period",			po::value<size_t>()->default_value(1000),						"[PT] Number of samples to produce during each temperature adaptation step.")
		("sampler.proposal_type",						po::value<std::string>()->default_value("autoblock"),			"[PT] Proposal type, can be either: scalar, vector, autoblock")
		("sampler.population_size",						po::value<size_t>()->default_value(1000),						"[SMC/Nested] Number of particles.")
		("sampler.start_temperature",					po::value<Real>()->default_value(0.0),							"[SMC] Starting temperature, between 0 and 1.")
		("sampler.end_temperature",						po::value<Real>()->default_value(1.0),							"[SMC] End temperature, between 0 and 1, typically 1.")
		("sampler.temperature_schedule",				po::value<std::string>()->default_value("adaptive"),			"[SMC] Schedule for the temperatures, can be \"fixed_linear\", \"fixed_loglinear\" or \"adaptive\".")
		("sampler.temperature_schedule_adaptive_ratio", po::value<Real>()->default_value(0.99),							"[SMC] Ratio specifying the target of the next population's ESS, [0-1). Only meaningful if temperature_schedule=adaptive")
		("sampler.temperature_schedule_fixed_number",	po::value<size_t>()->default_value(1000),						"[SMC] The number of iterations in a fixed temperature schedule. Only meaningful if temperature_schedule=fixed_linear or fixed_loglinear.")
		("sampler.resampling_threshold",				po::value<Real>()->default_value(0.5),							"[SMC] Threshold for resampling; the population will be resampled if ESS < threshold * population_size.")
		("sampler.smc_strategy",						po::value<std::string>()->default_value("mcmc_global"),			"[SMC] Strategy for sampling the next population, can be \"mcmc_global\", \"mcmc_global_covariance\", \"mcmc_local\", \"kde_local\".")
		("sampler.smc_mcmc_steps",						po::value<size_t>()->default_value(1),							"[SMC] Number of MCMC steps per population iteration. Only meaningful if smc_strategy=mcmc_global or mcmc_local.")
		("sampler.smc_mcmc_recalculate",				po::value<size_t>()->default_value(20),							"[SMC] After how many iterations to recalculate the proposal distribution. Only meaningful if smc_strategy=mcmc_global or mcmc_local.")
		("sampler.smc_mcmc_local_num_nearby",			po::value<size_t>()->default_value(2),							"[SMC] Number of nearby particles to calculate local proposal as a multiple of the dimensionality. Only meaningful if smc_strategy=mcmc_local.")
		("sampler.smc_mcmc_max_correlation",			po::value<Real>()->default_value(1.0),							"[SMC] Run MCMC until the correlation of all marginal parameters between successive steps is less than or equal to this threshold, in chunks of smc_mcmc_steps.")
		("sampler.smc_final_ess_threshold",				po::value<Real>()->default_value(0.0),							"[SMC] ")
		("sampler.smc_run_until_decorrelation",			po::value<bool>()->default_value(false),						"[SMC] ")
		("sampler.sample_weights",						po::value<bool>()->default_value("true"),						"[Nested] Specify whether to calculate the weights by sampling (\"true\") or by expectation value (\"false\"); sampling the weights allows an estimation of the uncertainty in the evidence approximation.")
		("sampler.termination_criterion",				po::value<std::string>()->default_value("remaining_evidence"),	"[Nested] Termination criterion, can be \"fixed\", \"ntimesh\" or \"remaining_evidence\".")
		("sampler.termination_num_steps",				po::value<size_t>()->default_value(1000),						"[Nested] Number of steps. Only meaningful if termination_criterion=fixed.")
		("sampler.termination_remaining_evidence",		po::value<Real>()->default_value(0.01),							"[Nested] Stop sampling when the current population can at most contribute this much to the evidence term. Only meaningful if termination_criterion=remaining_evidence.")
		("sampler.termination_ntimesh_factor",			po::value<Real>()->default_value(2.0),							"[Nested] Stop sampling when the number of steps exceeds this factor * population_size * H. Only used if termination_criterion=ntimesh.")
		("sampler.nested_sampling_strategy",			po::value<std::string>()->default_value("multinest"),			"[Nested] Strategy to use for proposing new samples, can be \"prior\", \"ellipsoid\", \"multinest\", or \"mcmc\".")
		("sampler.nested_sampling_strategy_mcmc_numsteps", po::value<size_t>()->default_value(100),						"[Nested] For MCMC sampling strategy, the number of MCMC steps to take for generating a new point.")
		("sampler.nested_num_posterior_samples",		po::value<size_t>()->default_value(1000),						"[Nested] Number of posterior samples to be generated, can be different from population_size")
		("sampler.is_proposal_gmm",						po::value<std::string>()->default_value(""),					"[IS] ")
		("sampler.rngseed",								po::value<unsigned long>()->default_value(0),					"Random number generator seed, set to 0 to use an arbitrary seed based on the computer time.")
		("sampler.multistart_state_file",				po::value<std::string>()->default_value("multistart.bin"),		"Filename for storing intermediate results for multistart runs.")
	;
}

void SamplerFactory::SerializeSampler(Sampler* sampler, boost::archive::binary_oarchive& ar)
{
	SamplerSMC* smc = dynamic_cast<SamplerSMC*>(sampler);
	if (smc) {
		ar << *smc;
		return;
	}
	SamplerPT* pt = dynamic_cast<SamplerPT*>(sampler);
	if (pt) {
		ar << *pt;
		return;
	}
	SamplerNested* nested = dynamic_cast<SamplerNested*>(sampler);
	if (nested) {
		ar << *nested;
		return;
	}
	SamplerMH* mh = dynamic_cast<SamplerMH*>(sampler);
	if (mh) {
		ar << *mh;
		return;
	}

	ar << *sampler;
}

void SamplerFactory::SerializeSampler(Sampler* sampler, boost::archive::binary_iarchive& ar)
{
	SamplerSMC* smc = dynamic_cast<SamplerSMC*>(sampler);
	if (smc) {
		ar >> *smc;
		return;
	}
	SamplerPT* pt = dynamic_cast<SamplerPT*>(sampler);
	if (pt) {
		ar >> *pt;
		return;
	}
	SamplerNested* nested = dynamic_cast<SamplerNested*>(sampler);
	if (nested) {
		ar >> *nested;
		return;
	}
	SamplerMH* mh = dynamic_cast<SamplerMH*>(sampler);
	if (mh) {
		ar >> *mh;
		return;
	}
	SamplerIS* is = dynamic_cast<SamplerIS*>(sampler);
	if (is) {
		ar >> *is;
		return;
	}

	ar >> *sampler;
}

}
