#include "Utils.h"
#include "SBMLRatelaws.h"

bool SBMLRatelawElement::Generate(const ASTNode* node, const std::map<std::string, size_t>& species_index_map, const std::map<std::string, size_t>& parameter_index_map, SBMLRatelawElement** element)
{
	switch (node->getType()) {
	case AST_PLUS:
		*element = new SBMLRatelawElementPlus;
		(*element)->children.resize(node->getNumChildren());
		for (unsigned int i = 0; i < node->getNumChildren(); i++) {
			if (!Generate(node->getChild(i), species_index_map, parameter_index_map, &(*element)->children[i])) {
				return false;
			}
		}
		break;

	case AST_MINUS:
		if (node->getNumChildren() == 1) {
			*element = new SBMLRatelawElementNegate;
			(*element)->children.resize(1);
			if (!Generate(node->getChild(0), species_index_map, parameter_index_map, &(*element)->children[0])) {
				return false;
			}
		} else if (node->getNumChildren() == 2) {
			*element = new SBMLRatelawElementMinus;
			(*element)->children.resize(2);
			if (!Generate(node->getChild(0), species_index_map, parameter_index_map, &(*element)->children[0])) {
				return false;
			}
			if (!Generate(node->getChild(1), species_index_map, parameter_index_map, &(*element)->children[1])) {
				return false;
			}
		} else {
			// I'm not sure how this is defined..
			LOGERROR("Not implemented");
			return false;
		}
		break;

	case AST_TIMES:
		*element = new SBMLRatelawElementTimes;
		(*element)->children.resize(node->getNumChildren());
		for (unsigned int i = 0; i < node->getNumChildren(); i++) {
			if (!Generate(node->getChild(i), species_index_map, parameter_index_map, &(*element)->children[i])) {
				return false;
			}
		}
		break;

	case AST_DIVIDE:
		if (node->getNumChildren() == 2) {
			*element = new SBMLRatelawElementDivide;
			(*element)->children.resize(2);
			if (!Generate(node->getChild(0), species_index_map, parameter_index_map, &(*element)->children[0])) {
				return false;
			}
			if (!Generate(node->getChild(1), species_index_map, parameter_index_map, &(*element)->children[1])) {
				return false;
			}
		} else {
			// I'm not sure how this is defined..
			LOGERROR("AST division with more than 2 children?");
			return false;
		}
		break;

	case AST_NAME:
		{
			const char* name = node->getName();
			std::map<std::string, size_t>::const_iterator iter;

			iter = parameter_index_map.find(name);
			if (iter != parameter_index_map.end()) {
				SBMLRatelawElementLookupParameter* el = new SBMLRatelawElementLookupParameter;
				el->ix = iter->second;
				*element = el;
				break;
			}

			iter = species_index_map.find(name);
			if (iter != species_index_map.end()) {
				SBMLRatelawElementLookupSpecies* el = new SBMLRatelawElementLookupSpecies;
				el->ix = iter->second;
				*element = el;
				break;
			}

			LOGERROR("AST_NAME name \"%s\" does not map to either a species id or a parameter", name);
			return false;
		}
		break;

	case AST_INTEGER:
		{
			SBMLRatelawElementConstant* el = new SBMLRatelawElementConstant;
			el->constant = node->getInteger();
			*element = el;
		}
		break;

	case AST_REAL:
		{
			SBMLRatelawElementConstant* el = new SBMLRatelawElementConstant;
			el->constant = node->getReal();
			*element = el;
		}
		break;

	case AST_FUNCTION_EXP:
		if (node->getNumChildren() == 1) {
			*element = new SBMLRatelawElementFunctionExp;
			(*element)->children.resize(1);
			if (!Generate(node->getChild(0), species_index_map, parameter_index_map, &(*element)->children[0])) {
				return false;
			}
		} else {
			LOGERROR("AST exp with wrong number of children?");
			return false;
		}
		break;

	case AST_FUNCTION_LN:
		if (node->getNumChildren() == 1) {
			*element = new SBMLRatelawElementFunctionLog;
			(*element)->children.resize(1);
			if (!Generate(node->getChild(0), species_index_map, parameter_index_map, &(*element)->children[0])) {
				return false;
			}
		} else {
			LOGERROR("AST log with wrong number of children?");
			return false;
		}
		break;

	case AST_FUNCTION_POWER:
		if (node->getNumChildren() == 2) {
			*element = new SBMLRatelawElementFunctionPow;
			(*element)->children.resize(2);
			if (!Generate(node->getChild(0), species_index_map, parameter_index_map, &(*element)->children[0])) {
				return false;
			}
			if (!Generate(node->getChild(1), species_index_map, parameter_index_map, &(*element)->children[1])) {
				return false;
			}
		} else {
			LOGERROR("AST pow with wrong number of children?");
			return false;
		}
		break;

	default:
		LOGERROR("SBML AST node type %d not implemented", (int)node->getType());
		return false;
	}

	return true;
}

SBMLRatelawElement::~SBMLRatelawElement()
{
	for (size_t i = 0; i < children.size(); i++) {
		delete children[i];
	}
}

bool SBMLRatelawElement::ContainsSpeciesLookup(size_t species_ix)
{
	for (size_t i = 0; i < children.size(); i++) {
		if (children[i]->ContainsSpeciesLookup(species_ix)) {
			return true;
		}
	}
	return false;
}

Real SBMLRatelawElementPlus::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() > 1);
	Real result = children[0]->Evaluate(species, parameters);
	for (size_t i = 1; i < children.size(); i++) {
		result += children[i]->Evaluate(species, parameters);
	}
	return result;
}

std::string SBMLRatelawElementPlus::GenerateEquation()
{
	ASSERT(children.size() > 1);
	std::string result = "(";
	result += children[0]->GenerateEquation();
	for (size_t i = 1; i < children.size(); i++) {
		result += "+";
		result += children[i]->GenerateEquation();
	}
	result += ")";
	return result;
}

std::string SBMLRatelawElementPlus::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() > 1);

	std::vector<std::string> children_result;
	bool nonzero = false;
	for (size_t i = 0; i < children.size(); i++) {
		std::string res = children[i]->GenerateDerivative(species_ix);
		children_result.push_back(res);
		if (!res.empty()) {
			nonzero = true;
		}
	}

	if (nonzero) {
		std::string result = "(";
		for (size_t i = 0; i < children.size(); i++) {
			if (!children_result[i].empty()) {
				result += "+";
				result += children_result[i];
			}
		}
		result += ")";
		return result;
	} else {
		return "";
	}
}

Real SBMLRatelawElementMinus::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() == 2);
	return children[0]->Evaluate(species, parameters) - children[1]->Evaluate(species, parameters);
}

std::string SBMLRatelawElementMinus::GenerateEquation()
{
	ASSERT(children.size() == 2);
	std::string result = "(";
	result += children[0]->GenerateEquation();
	result += "-";
	result += children[1]->GenerateEquation();
	result += ")";
	return result;
}

std::string SBMLRatelawElementMinus::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() == 2);

	std::string res1 = children[0]->GenerateDerivative(species_ix);
	std::string res2 = children[1]->GenerateDerivative(species_ix);

	if (res1.empty()) {
		if (res2.empty()) {
			return "";
		} else {
			std::string result = "(";
			result += "-";
			result += res2;
			result += ")";
			return result;
		}
	} else {
		if (res2.empty()) {
			return res1;
		} else {
			std::string result = "(";
			result += res1;
			result += "-";
			result += res2;
			result += ")";
			return result;
		}
	}
}

Real SBMLRatelawElementNegate::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() == 1);
	return -children[0]->Evaluate(species, parameters);
}

std::string SBMLRatelawElementNegate::GenerateEquation()
{
	ASSERT(children.size() == 1);
	std::string result = "(";
	result += "-";
	result += children[0]->GenerateEquation();
	result += ")";
	return result;
}

std::string SBMLRatelawElementNegate::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() == 1);
	std::string res = children[0]->GenerateDerivative(species_ix);
	if (res.empty()) {
		return "";
	} else {
		std::string result = "(";
		result += "-";
		result += res;
		result += ")";
		return result;
	}
}

Real SBMLRatelawElementTimes::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() > 1);
	Real result = children[0]->Evaluate(species, parameters);
	for (size_t i = 1; i < children.size(); i++) {
		result *= children[i]->Evaluate(species, parameters);
	}
	return result;
}

std::string SBMLRatelawElementTimes::GenerateEquation()
{
	ASSERT(children.size() > 1);
	std::string result = "(";
	result += children[0]->GenerateEquation();
	for (size_t i = 1; i < children.size(); i++) {
		result += "*";
		result += children[i]->GenerateEquation();
	}
	result += ")";
	return result;
}

std::string SBMLRatelawElementTimes::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() > 1);

	std::vector<std::string> children_result;
	bool nonzero = false;
	for (size_t i = 0; i < children.size(); i++) {
		std::string res = children[i]->GenerateDerivative(species_ix);
		children_result.push_back(res);
		if (!res.empty()) {
			nonzero = true;
		}
	}

	if (nonzero) {
		// Product rule
		std::string result = "(";
		for (size_t i = 0; i < children.size(); i++) {
			if (!children_result[i].empty()) {
				for (size_t j = 0; j < children.size(); j++) {
					if (j != 0) {
						result += "*";
					}
					if (i == j) {
						result += children_result[i];
					} else {
						result += children[j]->GenerateEquation();
					}
				}
			}
		}
		result += ")";
		return result;
	} else {
		return "";
	}
}

Real SBMLRatelawElementDivide::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() == 2);
	return children[0]->Evaluate(species, parameters) / children[1]->Evaluate(species, parameters);
}

std::string SBMLRatelawElementDivide::GenerateEquation()
{
	ASSERT(children.size() == 2);
	std::string result = "(";
	result += children[0]->GenerateEquation();
	result += "/";
	result += children[1]->GenerateEquation();
	result += ")";
	return result;
}

std::string SBMLRatelawElementDivide::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() == 2);

	std::string res1 = children[0]->GenerateDerivative(species_ix);
	std::string res2 = children[1]->GenerateDerivative(species_ix);
	std::string eq1 = children[0]->GenerateEquation();
	std::string eq2 = children[1]->GenerateEquation();

	if (res1.empty()) {
		if (res2.empty()) {
			return "";
		} else {
			std::string result = "(-";
			result += res2;
			result += "*";
			result += eq1;
			result += "/";
			result += "(";
			result += eq2;
			result += "*";
			result += eq2;
			result += ")";
			result += ")";
			return result;
		}
	} else {
		if (res2.empty()) {
			std::string result = "(";
			result += res1;
			result += "/";
			result += eq2;
			result += ")";
			return result;
		} else {
			std::string result = "(";
			result += "(";
			result += res1;
			result += "*";
			result += eq2;
			result += "-";
			result += res2;
			result += "*";
			result += eq1;
			result += ")";
			result += "/";
			result += "(";
			result += eq2;
			result += "*";
			result += eq2;
			result += ")";
			result += ")";
			return result;
		}
	}
}

Real SBMLRatelawElementLookupSpecies::Evaluate(const Real* species, const Real* parameters)
{
	return species[ix];
}

std::string SBMLRatelawElementLookupSpecies::GenerateEquation()
{
	std::string result = "species[";
	result += std::to_string((uint64)ix);
	result += "]";
	return result;
}

bool SBMLRatelawElementLookupSpecies::ContainsSpeciesLookup(size_t species_ix)
{
	return species_ix == ix;
}

std::string SBMLRatelawElementLookupSpecies::GenerateDerivative(size_t species_ix)
{
	if (species_ix == ix) {
		return "(OdeReal)1.0";
	} else {
		return "";
	}
}

Real SBMLRatelawElementLookupParameter::Evaluate(const Real* species, const Real* parameters)
{
	//return pow(10.0, parameters[ix]);
	return parameters[ix];
}

std::string SBMLRatelawElementLookupParameter::GenerateEquation()
{
	std::string result = "parameters[";
	result += std::to_string((uint64)ix);
	result += "]";
	return result;
}

bool SBMLRatelawElementLookupParameter::ContainsSpeciesLookup(size_t species_ix)
{
	return false;
}

std::string SBMLRatelawElementLookupParameter::GenerateDerivative(size_t species_ix)
{
	return "";
}

Real SBMLRatelawElementConstant::Evaluate(const Real* species, const Real* parameters)
{
	return constant;
}

std::string SBMLRatelawElementConstant::GenerateEquation()
{
	return std::to_string((long double)constant);
}

bool SBMLRatelawElementConstant::ContainsSpeciesLookup(size_t species_ix)
{
	return false;
}

std::string SBMLRatelawElementConstant::GenerateDerivative(size_t species_ix)
{
	return "";
}

Real SBMLRatelawElementFunctionExp::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() == 1);
	return exp(children[0]->Evaluate(species, parameters));
}

std::string SBMLRatelawElementFunctionExp::GenerateEquation()
{
	ASSERT(children.size() == 1);
	std::string result = "exp(";
	result += children[0]->GenerateEquation();
	result += ")";
	return result;
}

std::string SBMLRatelawElementFunctionExp::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() == 1);
	std::string res = children[0]->GenerateDerivative(species_ix);
	if (res.empty()) {
		return "";
	} else {
		std::string result = "exp(";
		result += children[0]->GenerateEquation();
		result += ")*(";
		result += res;
		result += ")";
		return result;
	}
}

Real SBMLRatelawElementFunctionLog::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() == 1);
	return exp(children[0]->Evaluate(species, parameters));
}

std::string SBMLRatelawElementFunctionLog::GenerateEquation()
{
	ASSERT(children.size() == 1);
	std::string result = "log(";
	result += children[0]->GenerateEquation();
	result += ")";
	return result;
}

std::string SBMLRatelawElementFunctionLog::GenerateDerivative(size_t species_ix)
{
	ASSERT(children.size() == 1);
	std::string res = children[0]->GenerateDerivative(species_ix);
	if (res.empty()) {
		return "";
	} else {
		std::string result = "(OdeReal)1.0/(";
		result += children[0]->GenerateEquation();
		result += ")";
		return result;
	}
}

Real SBMLRatelawElementFunctionPow::Evaluate(const Real* species, const Real* parameters)
{
	ASSERT(children.size() == 2);
	return pow(children[0]->Evaluate(species, parameters), children[1]->Evaluate(species, parameters));
}

std::string SBMLRatelawElementFunctionPow::GenerateEquation()
{
	std::string result = "pow(";
	result += children[0]->GenerateEquation();
	result += ",";
	result += children[1]->GenerateEquation();
	result += ")";
	return result;
}

std::string SBMLRatelawElementFunctionPow::GenerateDerivative(size_t species_ix)
{
	LOGERROR("Not implemented");
	ASSERT(false);
	// TODO.
	return "";
}
