#include "Utils.h"
#include "SBMLModel.h"
#include "SBMLRatelaws.h"
#include "SBMLReaction.h"
#include "SBMLSpecies.h"

SBMLReaction::SBMLReaction()
	: type(Invalid)
	, rate_law(NULL)
	, evaluate_rate_law(NULL)
{
}

SBMLReaction::~SBMLReaction()
{
	bcm::safe_delete(rate_law);
	bcm::safe_delete(evaluate_rate_law);
}

bool SBMLReaction::Initialize(const Reaction* reaction, const SBMLModel* model)
{
	id = reaction->getId();

	rate_law = reaction->getKineticLaw()->getMath()->deepCopy();

	const XMLNode* annotation = reaction->getAnnotation();
	if (annotation) {
		for (unsigned int i = 0; i < annotation->getNumChildren(); i++) {
			const XMLNode& child = annotation->getChild(i);
			if (child.getPrefix() == "celldesigner" && child.getName() == "extension") {
				std::string reaction_type_str = child.getChild("reactionType").getChild(0).getCharacters();
				if (reaction_type_str.empty()) {
					LOGERROR("Can't find reaction type for reaction %s - is the SBML file a CellDesigner file?", id.c_str());
					return false;
				}
			
				if (reaction_type_str == "HETERODIMER_ASSOCIATION") {
					type = Association;
				} else if (reaction_type_str == "DISSOCIATION") {
					type = Dissociation;
				} else if (reaction_type_str == "STATE_TRANSITION") {
					type = Transition;
				} else if (reaction_type_str == "TRANSCRIPTION") {
					type = Transcription;
				} else if (reaction_type_str == "TRANSLATION") {
					type = Translation;
				} else if (reaction_type_str == "PHYSICAL_STIMULATION") {
					type = PhysicalStimulation;
				} else {
					LOGERROR("Unrecognized reaction type %s for reaction %s", reaction_type_str.c_str(), id.c_str());
					return false;
				}
			}
		}
	}
	
	for (unsigned int i = 0; i < reaction->getNumReactants(); i++) {
		const std::string& species_id = reaction->getReactant(i)->getSpecies();
		const SBMLSpecies* species = model->GetSpecies(species_id);
		if (species && species->GetType() != SBMLSpecies::Sink) {
			reactants.push_back(species_id);
		}
	}

	for (unsigned int i = 0; i < reaction->getNumProducts(); i++) {
		const std::string& species_id = reaction->getProduct(i)->getSpecies();
		const SBMLSpecies* species = model->GetSpecies(species_id);
		if (species && species->GetType() != SBMLSpecies::Sink) {
			products.push_back(species_id);
		}
	}

	return true;
}

bool SBMLReaction::SetSpeciesVector(const std::vector<std::string>& species_ids)
{
	// Check if any of the named AST nodes map to a species
	MapRateLawSpecies(species_ids, rate_law);

	// Map reactants and products
	for (std::vector<std::string>::iterator ri = reactants.begin(); ri != reactants.end(); ++ri) {
		bool found = false;
		for (size_t i = 0; i < species_ids.size(); i++) {
			if (species_ids[i] == *ri) {
				reactant_indices.push_back(i);
				found = true;
				break;
			}
		}

		if (!found) {
			LOGERROR("Can't find reactant \"%s\" in the species vector for reaction %s", ri->c_str(), id.c_str());
			return false;
		}
	}
	
	for (std::vector<std::string>::iterator pi = products.begin(); pi != products.end(); ++pi) {
		bool found = false;
		for (size_t i = 0; i < species_ids.size(); i++) {
			if (species_ids[i] == *pi) {
				product_indices.push_back(i);
				found = true;
				break;
			}
		}

		if (!found) {
			LOGERROR("Can't find product %s in the species vector", pi->c_str());
			return false;
		}
	}

	return true;
}

bool SBMLReaction::SetVariableSet(const bcm::VariableSet* variables)
{
	// Check if any of the named AST nodes map to a variable
	MapRateLawParameters(variables, rate_law);
	return true;
}

bool SBMLReaction::PostInitialize()
{
	bcm::safe_delete(evaluate_rate_law);
	return SBMLRatelawElement::Generate(rate_law, species_index_map, parameter_index_map, &evaluate_rate_law);
}

bool SBMLReaction::AddRates(const Real* species, const Real* parameters, Real* rates)
{
	Real r = evaluate_rate_law->Evaluate(species, parameters);
	for (std::vector<size_t>::iterator ri = reactant_indices.begin(); ri != reactant_indices.end(); ++ri) {
		rates[*ri] -= r;
	}
	for (std::vector<size_t>::iterator pi = product_indices.begin(); pi != product_indices.end(); ++pi) {
		rates[*pi] += r;
	}
	return true;
}

void SBMLReaction::AddRateEquation(size_t species_ix, std::string& eqn)
{
	for (std::vector<size_t>::iterator ri = reactant_indices.begin(); ri != reactant_indices.end(); ++ri) {
		if (*ri == species_ix) {
			eqn += "-";
			eqn += evaluate_rate_law->GenerateEquation();
		}
	}
	for (std::vector<size_t>::iterator pi = product_indices.begin(); pi != product_indices.end(); ++pi) {
		if (*pi == species_ix) {
			eqn += "+";
			eqn += evaluate_rate_law->GenerateEquation();
		}
	}
}

void SBMLReaction::AddRateDerivativeEquation(size_t species_ix, size_t deriv_ix, std::string& eqn)
{
	for (std::vector<size_t>::iterator ri = reactant_indices.begin(); ri != reactant_indices.end(); ++ri) {
		if (*ri == species_ix) {
			std::string derivative = evaluate_rate_law->GenerateDerivative(deriv_ix);
			if (!derivative.empty()) {
				eqn += "-";
				eqn += derivative;
			}
		}
	}
	for (std::vector<size_t>::iterator pi = product_indices.begin(); pi != product_indices.end(); ++pi) {
		if (*pi == species_ix) {
			std::string derivative = evaluate_rate_law->GenerateDerivative(deriv_ix);
			if (!derivative.empty()) {
				eqn += "+";
				eqn += derivative;
			}
		}
	}
}

void SBMLReaction::MapRateLawSpecies(const std::vector<std::string>& species_ids, const ASTNode* node)
{
	if (node->getType() == AST_NAME) {
		const char* name = node->getName();
		for (size_t i = 0; i < species_ids.size(); i++) {
			if (species_ids[i] == name) {
				species_index_map[name] = i;
			}
		}
	} else {
		for (unsigned int i = 0; i < node->getNumChildren(); i++) {
			MapRateLawSpecies(species_ids, node->getChild(i));
		}
	}
}

void SBMLReaction::MapRateLawParameters(const bcm::VariableSet* variables, const ASTNode* node)
{
	if (node->getType() == AST_NAME) {
		const char* name = node->getName();
		for (size_t i = 0; i < variables->GetNumVariables(); i++) {
			if (variables->GetVariableName(i) == name) {
				parameter_index_map[name] = i;
			}
		}
	} else {
		for (unsigned int i = 0; i < node->getNumChildren(); i++) {
			MapRateLawParameters(variables, node->getChild(i));
		}
	}
}
