#include "Utils.h"
#include "ExperimentalCondition.h"
#include "LikelihoodExperiment.h"
#include "LikelihoodPart.h"
#include "LikelihoodInitialValue.h"
#include "LikelihoodIntegratedTrajectory.h"
#include "LikelihoodSingleTimepoint.h"
#include "LikelihoodTimeAverage.h"
#include "LikelihoodTrajectory.h"
#include "SBMLModel.h"

LikelihoodExperiment::LikelihoodExperiment(size_t numthreads)
	: num_threads(numthreads)
{
}

LikelihoodExperiment::~LikelihoodExperiment()
{
	for (std::vector<ExperimentalCondition*>::iterator ci = Conditions.begin(); ci != Conditions.end(); ++ci) {
		delete *ci;
	}
	for (std::vector<LikelihoodPart*>::iterator pi = Parts.begin(); pi != Parts.end(); ++pi) {
		delete *pi;
	}
}

bool LikelihoodExperiment::Load(const boost::property_tree::ptree& xml_node, const bcm::NetCDFDataFile* datafile)
{
	try {
		Name = xml_node.get<std::string>("<xmlattr>.name");

		std::string save_output = xml_node.get<std::string>("<xmlattr>.save_output", "__all__");
		bcm::tokenize(save_output, SaveOutputs, ";");

		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "condition") {
				ExperimentalCondition* condition = new ExperimentalCondition;
				if (!condition->Load(var.second, datafile)) {
					return false;
				}
				Conditions.push_back(condition);
			}
			if (var.first == "data") {
				LikelihoodPart* part = NULL;

				std::string type = var.second.get<std::string>("<xmlattr>.type");
				if (type == "trajectory") {
					part = new LikelihoodTrajectory;
				} else if (type == "single_timepoint") {
					part = new LikelihoodSingleTimepoint;
				} else if (type == "time_average") {
					part = new LikelihoodTimeAverage;
				} else if (type == "integrated_trajectory") {
					part = new LikelihoodIntegratedTrajectory;
				} else {
					LOGERROR("Unknown likelihood data type \"%s\"", type.c_str());
					return false;
				}

				if (!part->Load(var.second, datafile)) {
					delete part;
					return false;
				}

				Parts.push_back(part);
			}
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	return true;
}

bool LikelihoodExperiment::SetVariableSet(const bcm::VariableSet* variables)
{
	for (std::vector<LikelihoodPart*>::iterator pi = Parts.begin(); pi != Parts.end(); ++pi) {
		if (!(*pi)->SetVariableSet(variables)) {
			return false;
		}
	}

	return true;
}

bool LikelihoodExperiment::Initialize()
{
	// Get the timepoints at which we need the model solution
	for (std::vector<LikelihoodPart*>::iterator pi = Parts.begin(); pi != Parts.end(); ++pi) {
		std::vector<Real> part_timepoints;
		if (!(*pi)->GetRequiredTimepoints(part_timepoints)) {
			return false;
		}
		RequiredTimepoints.insert(RequiredTimepoints.end(), part_timepoints.begin(), part_timepoints.end());
	}
	std::sort(RequiredTimepoints.begin(), RequiredTimepoints.end());
	RequiredTimepoints.erase(std::unique(RequiredTimepoints.begin(), RequiredTimepoints.end()), RequiredTimepoints.end());

#if 1
	std::vector<Real> inflated;
	inflated.resize((RequiredTimepoints.size() * 4) - 3);
	for (size_t i = 0; i < RequiredTimepoints.size() - 1; i++) {
		inflated[i*4] = RequiredTimepoints[i];
		inflated[i*4+1] = 0.75 * RequiredTimepoints[i] + 0.25 * RequiredTimepoints[i+1];
		inflated[i*4+2] = 0.5  * RequiredTimepoints[i] + 0.5  * RequiredTimepoints[i+1];
		inflated[i*4+3] = 0.25 * RequiredTimepoints[i] + 0.75 * RequiredTimepoints[i+1];
	}
	inflated.back() = RequiredTimepoints.back();
	RequiredTimepoints = inflated;
#endif

	return true;
}

void LikelihoodExperiment::SetNumModelSpecies(size_t n)
{
	SimulatedTrajectories.resize(num_threads);
	for (size_t i = 0; i < num_threads; i++) {
		SimulatedTrajectories[i].resize(n);
	}
}

bool LikelihoodExperiment::SetupAdditionalOutput(bcm::NetCDFDataFile* file, SBMLModel* model, size_t sample_count)
{
	if (SaveOutputs.empty()) {
		return true;
	}

	//std::string group_name = std::string("simulated_output/") + Name;
	std::string group_name = Name;

	bool result = true;
	result &= file->CreateGroup(group_name);

	std::vector<unsigned int> sampleix(sample_count);
	for (size_t i = 0; i < sample_count; i++) {
		sampleix[i] = (unsigned int)i+1;
	}

	std::vector<std::string> species;
	model->GetSimulatedSpeciesFullNames(species);
	
	if (SaveOutputs.size() == 1 && SaveOutputs[0] == "__all__") {
		SaveOutputs = species;
	} else {
		for (std::vector<std::string>::iterator soi = SaveOutputs.begin(); soi != SaveOutputs.end(); ++soi) {
			if (std::find(species.begin(), species.end(), *soi) == species.end()) {
				LOGERROR("Unknown species %s in output save", soi->c_str());
				return false;
			}
		}
	}

	// TODO -actually restrict to the correct species

	result &= file->CreateDimension(group_name, "sample_ix", sampleix);
	result &= file->CreateDimension(group_name, "species", species);
	result &= file->CreateDimension(group_name, "time", RequiredTimepoints);
	result &= file->CreateVariable(group_name, "trajectories", "sample_ix", "species", "time");

	return result;
}

void LikelihoodExperiment::HandleAdditionalOutput(bcm::NetCDFDataFile* file, SBMLModel* model, size_t sample_ix, const Real* values, size_t& value_ix)
{
	if (SaveOutputs.empty()) {
		return;
	}

	//std::string group_name = std::string("simulated_output/") + Name;
	std::string group_name = Name;

	// TODO -actually restrict to the correct species
	
	file->PutValue(group_name, "sample_ix", sample_ix, (unsigned int)sample_ix);
	std::vector<size_t> ix(3), count(3);
	ix[0] = sample_ix;
	ix[1] = 0;
	ix[2] = 0;
	count[0] = 1;
	count[1] = SimulatedTrajectories[0].size();
	count[2] = RequiredTimepoints.size();
	file->PutValues(group_name, "trajectories", ix, count, values);

	// Update the index in the output array
	value_ix += RequiredTimepoints.size() * SimulatedTrajectories[0].size();
}

bool LikelihoodExperiment::EvaluateLogProbability(size_t threadix, const Real* values, SBMLModel* model, Real& logp)
{
	model->ResetForcedVariables(threadix);
	for (std::vector<ExperimentalCondition*>::iterator ci = Conditions.begin(); ci != Conditions.end(); ++ci) {
		if (!model->SetForcedVariable((*ci)->GetVariableName(), (*ci)->GetValue(), threadix)) {
			return false;
		}
	}

	if (!model->Simulate(values, RequiredTimepoints, threadix)) {
		logp = -std::numeric_limits<Real>::infinity();
	} else {
		logp = 0.0;
		for (std::vector<LikelihoodPart*>::iterator pi = Parts.begin(); pi != Parts.end(); ++pi) {
			Real part_logp = std::numeric_limits<Real>::quiet_NaN();
			if (!(*pi)->EvaluateLogProbability(threadix, values, model, RequiredTimepoints, part_logp)) {
				return false;
			}
			logp += part_logp;
		}
	}
	
	for (size_t i = 0; i < SimulatedTrajectories[threadix].size(); i++) {
		model->GetOutput(i, SimulatedTrajectories[threadix][i], threadix);
	}

	return true;
}

bool LikelihoodExperiment::GetOutput(size_t threadix, Real* output_values, size_t& numvalues)
{
	if (output_values) {
		for (size_t si = 0; si < SimulatedTrajectories[threadix].size(); si++) {
			if (SimulatedTrajectories[threadix][si].size() == RequiredTimepoints.size()) {
				for (size_t ti = 0; ti < RequiredTimepoints.size(); ti++) {
					output_values[si * RequiredTimepoints.size() + ti] = SimulatedTrajectories[threadix][si](ti);
				}
			} else {
				for (size_t ti = 0; ti < RequiredTimepoints.size(); ti++) {
					output_values[si * RequiredTimepoints.size() + ti] = std::numeric_limits<Real>::quiet_NaN();
				}
			}
		}
		numvalues += RequiredTimepoints.size() * SimulatedTrajectories[threadix].size();
	} else {
		numvalues += RequiredTimepoints.size() * SimulatedTrajectories[threadix].size();
	}

	return true;
}
