#include "Utils.h"
#include "SBMLPDExperiment.h"
#include "SBMLPDExperimentDrugRange.h"
#include "SBMLPDExperimentSingleCondition.h"
#include "SignalingNetwork.h"
#include "ProbabilityDistributions.h"

SBMLPDExperiment::DataPartBase::DataPartBase()
	: use_base(false)
	, use_scale(false)
	, fixed_var(false)
	, weight(1.0)
	, model_ix(std::numeric_limits<size_t>::max())
	, parameter_base_ix(std::numeric_limits<size_t>::max())
	, parameter_scale_ix(std::numeric_limits<size_t>::max())
	, parameter_variance_ix(std::numeric_limits<size_t>::max())
	, fixed_variance_value(0.0)
	, data_is_inactive_form(false)
	, model_ix2(std::numeric_limits<size_t>::max())
	, data_sum_parameter_ix(std::numeric_limits<size_t>::max())
{
}

SBMLPDExperiment::SBMLPDExperiment(size_t numthreads)
	: numthreads(numthreads)
	, network(NULL)
	, varset(NULL)
{
}

SBMLPDExperiment::~SBMLPDExperiment()
{
}

SBMLPDExperiment* SBMLPDExperiment::Create(const boost::property_tree::ptree& xml_node, SignalingNetwork* network, const bcm::VariableSet* varset, std::vector<SBMLPDExperiment*> other_experiments, size_t numthreads)
{
	SBMLPDExperiment* experiment = NULL;

	try {
		std::string type = xml_node.get<std::string>("<xmlattr>.type", "single_condition");

		if (type == "single_condition") {
			experiment = new SBMLPDExperimentSingleCondition(numthreads);
		} else if (type == "drug_range") {
			experiment = new SBMLPDExperimentDrugRange(numthreads);
		} else {
			LOGERROR("Unknown experiment type \"%s\"", type.c_str());
			return NULL;
		}

		experiment->Name = xml_node.get<std::string>("<xmlattr>.name");
		experiment->DataFile = xml_node.get<std::string>("<xmlattr>.datafile");
		experiment->network = network;
		experiment->varset = varset;

		if (!experiment->Load(xml_node, other_experiments)) {
			delete experiment;
			return NULL;
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		bcm::safe_delete(experiment);
		return NULL;
	}

	if (!experiment->InitializeParallelData()) {
		delete experiment;
		return NULL;
	}

	return experiment;
}

bool SBMLPDExperiment::Load(const boost::property_tree::ptree& xml_node, std::vector<SBMLPDExperiment*> other_experiments)
{
	bool result = true;
	bcm::NetCDFDataFile file;
	
	if (!file.Open(DataFile, netCDF::NcFile::read)) {
		return false;
	}

	size_t num_cell_lines;
	result &= file.GetDimensionSize(Name, "cell_lines", &num_cell_lines);
	cell_lines.resize(num_cell_lines);
	for (size_t i = 0; i < cell_lines.size(); i++) {
		result &= file.GetValue(Name, "cell_lines", i, cell_lines[i]);
	}

	if (!LoadTypeSpecificNodes(xml_node, &file)) {
		return false;
	}
	
	try {
		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "data") {
				if (!ParseDataNode(var.second, other_experiments, &file)) {
					return false;
				}
			} else if (var.first == "mutation" || var.first == "condition") {
				ModelConditions c;
				c.model_name = var.second.get<std::string>("<xmlattr>.species_name");
				c.model_ix = network->GetSignalingMoleculeIxByName(c.model_name);
				if (c.model_ix == std::numeric_limits<size_t>::max()) {
					LOGERROR("Could not find signaling node \"%s\" for setting model condition", c.model_name.c_str());
					return false;
				}

				std::string value_str = var.second.get<std::string>("<xmlattr>.value", "");
				std::string data_name = var.second.get<std::string>("<xmlattr>.data_name", "");

				if (!data_name.empty()) {
					std::string data_name_without_index;
					std::vector<size_t> data_indices;
					if (!ParseDataFileReference(data_name, data_name_without_index, data_indices, &file)) {
						LOGERROR("Unable to parse data reference \"%s\" for model condition for species \"%s\"", data_name.c_str(), c.model_name.c_str());
						return false;
					}
					if (data_indices.size() != 2) {
						LOGERROR("Model condition for species \"%s\" specifies a data reference \"%s\", but it does not have 2 dimensions (should be: [genes/epitopes,cell_lines])", c.model_name.c_str(), data_name.c_str());
						return false;
					}

					// TODO - verify dimension 2 is cell lines
					
					result &= file.GetValuesDim2(Name, data_name_without_index, data_indices[0], 0, cell_lines.size(), c.values);
				} else if (!value_str.empty()) {
					Real value = boost::lexical_cast<Real>(value_str);
					c.values = VectorReal::Constant(cell_lines.size(), value);
				} else {
					LOGERROR("Model condition for variable \"%s\" has neither data_name nor value specified", c.model_name.c_str());
					return false;
				}

				conditions.push_back(c);
			} else if (var.first == "expression_level") {
				ExpressionLevel el;
				el.model_name = var.second.get<std::string>("<xmlattr>.species_name");
				el.model_ix = network->GetSignalingMoleculeIxByName(el.model_name);
				if (el.model_ix == std::numeric_limits<size_t>::max()) {
					LOGERROR("Could not find signaling node \"%s\" for setting expression level", el.model_name.c_str());
					return false;
				}

				std::string value_str = var.second.get<std::string>("<xmlattr>.value", "");
				std::string data_name = var.second.get<std::string>("<xmlattr>.data_name", "");
				
				if (!data_name.empty()) {
					std::string data_name_without_index;
					std::vector<size_t> data_indices;
					if (!ParseDataFileReference(data_name, data_name_without_index, data_indices, &file)) {
						LOGERROR("Unable to parse data reference \"%s\" for expression level for species \"%s\"", data_name.c_str(), el.model_name.c_str());
						return false;
					}
					if (data_indices.size() != 2) {
						LOGERROR("Expression level for species \"%s\" specifies a data reference \"%s\", but it does not have 2 dimensions (should be: [genes/epitopes,cell_lines])", el.model_name.c_str(), data_name.c_str());
						return false;
					}

					// TODO - verify dimension 2 is cell lines
					
					result &= file.GetValuesDim2(Name, data_name_without_index, data_indices[0], 0, cell_lines.size(), el.values);
				} else if (!value_str.empty()) {
					Real value = boost::lexical_cast<Real>(value_str);
					el.values = VectorReal::Constant(cell_lines.size(), value);
				} else {
					LOGERROR("Expression level for variable \"%s\" has neither data_name nor value specified", el.model_name.c_str());
					return false;
				}
				
				std::string parameter_str = var.second.get<std::string>("<xmlattr>.parameter", std::string("expression[") + el.model_name + std::string("]"));
				el.parameter_ix = varset->GetVariableIndex(parameter_str, false);
				
				expression_levels.push_back(el);
			}
		}
		
		// Legacy - alternative means of specifying model conditions
		if (file.VariableExists(Name, "model_conditions_nodes")) {
			size_t num_conditions;
			result &= file.GetDimensionSize(Name, "model_conditions_nodes", &num_conditions);
			for (size_t i = 0; i < num_conditions; i++) {
				ModelConditions c;
				result &= file.GetValue(Name, "model_conditions_nodes", i, c.model_name);
				c.values = VectorReal::Constant(cell_lines.size(), std::numeric_limits<Real>::quiet_NaN());
				result &= file.GetValuesDim2(Name, "model_conditions", i, 0, cell_lines.size(), c.values);

				c.model_ix = network->GetSignalingMoleculeIxByName(c.model_name);
				if (c.model_ix == std::numeric_limits<size_t>::max()) {
					LOGERROR("Could not find signaling node \"%s\" for setting model condition", c.model_name.c_str());
					return false;
				}
				conditions.push_back(c);
			}
		} else {
		}
		
		// Legacy - alternative means of specifying expression levels
		if (file.VariableExists(Name, "expression_level_nodes")) {
			size_t num_expression_levels;
			result &= file.GetDimensionSize(Name, "expression_level_nodes", &num_expression_levels);
			for (size_t i = 0; i < num_expression_levels; i++) {
				ExpressionLevel& el = expression_levels[i];

				result &= file.GetValue(Name, "expression_level_nodes", i, el.model_name);
				el.model_ix = network->GetSignalingMoleculeIxByName(el.model_name);
				if (el.model_ix == std::numeric_limits<size_t>::max()) {
					LOGERROR("Could not find signaling node \"%s\" for setting expression level", el.model_name.c_str());
					return false;
				}

				VectorReal values = VectorReal::Constant(cell_lines.size(), std::numeric_limits<Real>::quiet_NaN());
				result &= file.GetValuesDim2(Name, "expression_levels", i, 0, cell_lines.size(), values);
				el.values = values;
				
				std::string parameter_str = std::string("expression[") + el.model_name + std::string("]");
				el.parameter_ix = varset->GetVariableIndex(parameter_str, false);
			}
		} else {
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	file.Close();
	return true;
}

bool SBMLPDExperiment::ParseDataPartBase(const boost::property_tree::ptree& xml_node, DataPartBase& p, std::vector<SBMLPDExperiment*> other_experiments)
{
	p.model_name = xml_node.get<std::string>("<xmlattr>.species_name");
	p.data_name  = xml_node.get<std::string>("<xmlattr>.data_name");
	p.base_scale_var_suffix  = xml_node.get<std::string>("<xmlattr>.base_scale_var_suffix", "");

	p.use_base = xml_node.get<bool>("<xmlattr>.use_base", true);
	p.use_scale = xml_node.get<bool>("<xmlattr>.use_scale", true);
	p.weight = xml_node.get<Real>("<xmlattr>.weight", 1.0);

	p.data_is_inactive_form = xml_node.get<bool>("<xmlattr>.data_is_inactive_form", false);

	if (p.use_base) {
		p.parameter_base_ix = varset->GetVariableIndex(std::string("base_") + p.base_scale_var_suffix);
		if (p.parameter_base_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Could not find variable for data base parameter with suffix \"%s\"", p.base_scale_var_suffix.c_str());
			return false;
		}
	}
				
	if (p.use_scale) {
		p.parameter_scale_ix = varset->GetVariableIndex(std::string("scale_") + p.base_scale_var_suffix);
		if (p.parameter_scale_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Could not find variable for data scale parameter with suffix \"%s\"", p.base_scale_var_suffix.c_str());
			return false;
		}
	}

	std::string variance_str = xml_node.get<std::string>("<xmlattr>.variance", "");
	if (variance_str.empty()) {
		variance_str = std::string("variance_") + p.base_scale_var_suffix;
	}

	// Is the variance a parameter in the prior?
	p.parameter_variance_ix = varset->GetVariableIndex(variance_str, false);
	if (p.parameter_variance_ix == std::numeric_limits<size_t>::max()) {
		// No, it isn't - try to cast it to a Real
		try {
			p.fixed_variance_value = boost::lexical_cast<Real>(variance_str);
			p.fixed_var = true;
		} catch (const boost::bad_lexical_cast& e) {
			LOGERROR("Could not find variable for data variance parameter \"%s\", and could also not cast it to a constant real value: %s", variance_str.c_str(), e.what());
			return false;
		}
	}
	else {
		p.fixed_var = false;
	}

	// Is the data a sum of two model variables?
	if (p.model_name.find('+') != std::string::npos) {
		std::vector<std::string> vars;
		bcm::tokenize(p.model_name, vars, "+");
		if (vars.size() != 2) {
			LOGERROR("Species name in data reference \"%s\" has a '+' sign, indicating the data is a sum of model species; but there are %z species and only 2 species are supported.", p.model_name.c_str(), vars.size());
			return false;
		}
		p.model_ix = network->GetSignalingMoleculeIxByName(vars[0]);
		if (p.model_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Could not find signaling node \"%s\"", vars[0].c_str());
			return false;
		}

		p.model_ix2 = network->GetSignalingMoleculeIxByName(vars[1]);
		if (p.model_ix2 == std::numeric_limits<size_t>::max()) {
			LOGERROR("Could not find signaling node \"%s\"", vars[1].c_str());
			return false;
		}

		std::string data_sum_parameter_str = xml_node.get<std::string>("<xmlattr>.data_sum_parameter");
		p.data_sum_parameter_ix = varset->GetVariableIndex(data_sum_parameter_str, false);
		if (p.data_sum_parameter_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Could not find variable for data sum parameter \"%s\"", data_sum_parameter_str.c_str());
			return false;
		}
	} else {
		p.model_ix = network->GetSignalingMoleculeIxByName(p.model_name);
		if (p.model_ix == std::numeric_limits<size_t>::max()) {
			LOGERROR("Could not find signaling node \"%s\"", p.model_name.c_str());
			return false;
		}

		p.model_ix2 = std::numeric_limits<size_t>::max();
		p.data_sum_parameter_ix = std::numeric_limits<size_t>::max();
	}

	std::string type = xml_node.get<std::string>("<xmlattr>.type", "");
	if (type == "relative") {
		p.relative_to_experiment = xml_node.get<std::string>("<xmlattr>.relative_reference");
					
		// Make sure the other experiment is defined before this one
		bool found = false;
		for (std::vector<SBMLPDExperiment*>::iterator ei = other_experiments.begin(); ei != other_experiments.end(); ++ei) {
			if ((*ei)->Name == p.relative_to_experiment) {
				found = true;
				break;
			}
		}
		if (!found) {
			LOGERROR("A data point for node \"%s\" is specified as relative to experiment \"%s\", but that experiment has not been defined. Make sure relative experiments are defined in the right order.", p.data_name.c_str(), p.relative_to_experiment.c_str());
			return false;
		}
	}

	return true;
}

void SBMLPDExperiment::PrepareActivitiesCalculation(VectorReal& activities, VectorReal& expression, const Real* transformed_values, size_t cell_ix)
{
	// Initialize activities to NaN, so that SignalingNetwork knows which values it still has to calculate.
	activities.setConstant(std::numeric_limits<Real>::quiet_NaN());

	// Clear all drug concentrations and set the new concentration for this drug
	network->SetDrugConcentrations(activities, 0.0);

	// Fill the activities for the nodes which are set to a value in the model conditions.
	for (size_t mci = 0; mci < conditions.size(); mci++) {
		const ModelConditions& c = conditions[mci];
		activities(c.model_ix) = c.values(cell_ix);
	}
	
	// Calculate expression values
	expression = VectorReal::Ones(network->GetMoleculeCount());
	for (std::vector<ExpressionLevel>::const_iterator eli = expression_levels.begin(); eli != expression_levels.end(); ++eli) {
		if (eli->parameter_ix != std::numeric_limits<size_t>::max()) {
			const Real e = transformed_values[eli->parameter_ix];
			expression(eli->model_ix) = e * eli->values(cell_ix) + (1.0 - e);
		} else {
			expression(eli->model_ix) = eli->values(cell_ix);
		}
	}
}

bool SBMLPDExperiment::ParseDataFileReference(const std::string& data_name, std::string& data_name_without_index, std::vector<size_t>& data_indices, const bcm::NetCDFDataFile* datafile)
{
	//
	// The data name parsing is copied from sbmlodeinf/LikelihoodPart.cpp
	//

	data_indices.clear();

	size_t indexing_start = data_name.find_first_of('[');
	if (indexing_start == data_name.npos) {
		// No indexing specified
		data_name_without_index = data_name;
		return true;
	}

	size_t indexing_end = data_name.find_first_of(']');
	if (indexing_end == data_name.npos) {
		LOGERROR("Syntax error in data name: missing closing bracket for indexing");
		return false;
	}

	data_name_without_index = data_name.substr(0, indexing_start);

	std::string indexingstr = data_name.substr(indexing_start+1, indexing_end - indexing_start - 1);
	std::vector<std::string> indexing_parts;
	bcm::tokenize(indexingstr, indexing_parts, ",");

	for (size_t i = 0; i < indexing_parts.size(); i++) {
		if (indexing_parts[i] == ":") {
			data_indices.push_back(std::numeric_limits<size_t>::max());
		} else {
			if (indexing_parts[i].front() == '\'' && indexing_parts[i].back() == '\'') {
				std::string val = indexing_parts[i].substr(1, indexing_parts[i].size() - 2);
				std::string dimname;
				if (!datafile->GetDimensionName(Name, data_name_without_index, i, dimname)) {
					return false;
				}

				size_t ix;
				if (!datafile->GetDimensionIx(Name, dimname, val, &ix)) {
					return false;
				}
				data_indices.push_back(ix);
			} else {
				size_t ix;
				try {
					ix = boost::lexical_cast<size_t>(indexing_parts[i]);
				} catch( boost::bad_lexical_cast const& ) {
					LOGERROR("Error converting index to a number");
					return false;
				}
				data_indices.push_back(ix);
			}
		}
	}

	return true;
}
